@@ -1,24 +1,32 @@
 #!/bin/bash
 . /usr/share/beakerlib/beakerlib.sh || exit 1
+. ../../images.sh || exit 1
+
+PROVISION_HOW=container
 
 rlJournalStart
     rlPhaseStartSetup
+
+        build_container_image "centos/7/upstream\:latest"
+        build_container_image "fedora/40:\latest"
+
         rlRun "pushd data"
         rlRun "set -o pipefail"
     rlPhaseEnd
 
-    for image in fedora centos:7 ; do
+    for image in $TEST_IMAGE_PREFIX/fedora/40:latest \
+                 $TEST_IMAGE_PREFIX/centos/7/upstream:latest; do
         # Prepare the tmt command and expected error message
-        tmt="tmt run -avr provision -h container -i $image"
-        if [[ $image == fedora ]]; then
-            error='Unable to find a match: forest'
+        tmt="tmt run -avr provision -h $PROVISION_HOW -i $image"
+        if [[ $image =~ fedora ]]; then
+            error='No match for argument: forest'
         else
             error='No package forest available'
         fi
 
         rlPhaseStartTest "Require an available package ($image)"
             rlRun -s "$tmt plan --name available"
-            rlAssertGrep '1 preparation applied' $rlRun_LOG
+            rlAssertGrep '2 preparations applied' $rlRun_LOG
         rlPhaseEnd
 
         rlPhaseStartTest "Require a missing package ($image)"
