@@ -24,7 +24,6 @@ but all-capturing log files while keeping implementation simple - the other opti
 managing handlers themselves, which would be very messy given the propagation of messages.
 """
 
-import dataclasses
 import enum
 import itertools
 import logging
@@ -37,15 +36,19 @@ from typing import (
     Callable,
     Optional,
     Protocol,
+    TextIO,
     Union,
     cast,
-    )
+)
 
-import click
+from tmt._compat.pathlib import Path
+from tmt._compat.warnings import deprecated
+from tmt.container import container, simple_field
 
 if TYPE_CHECKING:
     import tmt.cli
     import tmt.utils
+    import tmt.utils.themes
 
 # Log in workdir
 LOG_FILENAME = 'log.txt'
@@ -73,6 +76,7 @@ LABEL_FORMAT = '[{label}]'
 
 LoggableValue = Union[
     str,
+    dict[str, Any],
     int,
     bool,
     float,
@@ -80,7 +84,8 @@ LoggableValue = Union[
     'tmt.utils.FmfContext',
     'tmt.utils.Path',
     'tmt.utils.Command',
-    'tmt.utils.ShellScript']
+    'tmt.utils.ShellScript',
+]
 
 
 # TODO: this is an ugly hack, removing colors after they have been added...
@@ -182,21 +187,24 @@ def render_labels(labels: list[str]) -> str:
     if not labels:
         return ''
 
+    from tmt.utils.themes import style
+
     return ''.join(
         # TODO: color here is questionable - it will be removed, but I'd rather not
         # add it at first place, and it should be configurable.
-        click.style(LABEL_FORMAT.format(label=label), fg='cyan')
+        style(LABEL_FORMAT.format(label=label), fg='cyan')
         for label in labels
-        )
+    )
 
 
 def indent(
-        key: str,
-        value: Optional[LoggableValue] = None,
-        color: Optional[str] = None,
-        level: int = 0,
-        labels: Optional[list[str]] = None,
-        labels_padding: int = 0) -> str:
+    key: str,
+    value: Optional[LoggableValue] = None,
+    color: 'tmt.utils.themes.Style' = None,
+    level: int = 0,
+    labels: Optional[list[str]] = None,
+    labels_padding: int = 0,
+) -> str:
     """
     Indent a key/value message.
 
@@ -215,11 +223,12 @@ def indent(
         length.
     """
 
+    from tmt.utils.themes import style
+
     indent = ' ' * INDENT * level
 
     # Colorize
-    if color is not None:
-        key = click.style(key, fg=color)
+    key = style(key, style=color)
 
     # Prepare prefix if labels provided
     prefix = render_labels(labels).ljust(labels_padding) + ' ' if labels else ''
@@ -245,21 +254,24 @@ def indent(
     # extra bit of indentation ("deeper").
     deeper = ' ' * INDENT
 
-    return f'{prefix}{indent}{key}:\n' \
-        + '\n'.join(f'{prefix}{indent}{deeper}{line}' for line in lines)
+    return f'{prefix}{indent}{key}:\n' + '\n'.join(
+        f'{prefix}{indent}{deeper}{line}' for line in lines
+    )
 
 
-@dataclasses.dataclass
+@container
 class LogRecordDetails:
-    """ tmt's log message components attached to log records """
+    """
+    tmt's log message components attached to log records
+    """
 
     key: str
     value: Optional[LoggableValue] = None
 
-    color: Optional[str] = None
+    color: 'tmt.utils.themes.Style' = None
     shift: int = 0
 
-    logger_labels: list[str] = dataclasses.field(default_factory=list)
+    logger_labels: list[str] = simple_field(default_factory=list)
     logger_labels_padding: int = 0
 
     logger_verbosity_level: int = 0
@@ -271,14 +283,19 @@ class LogRecordDetails:
     logger_quiet: bool = False
     ignore_quietness: bool = False
 
-    logger_topics: set[Topic] = dataclasses.field(default_factory=set)
+    logger_topics: set[Topic] = simple_field(default_factory=set)
     message_topic: Optional[Topic] = None
 
 
 class LogfileHandler(logging.FileHandler):
+    #: Paths of all log files to which ``LogfileHandler`` was attached.
+    emitting_to: list[Path] = []
+
     def __init__(self, filepath: 'tmt.utils.Path') -> None:
         super().__init__(filepath, mode='a')
 
+        LogfileHandler.emitting_to.append(filepath)
+
 
 # ignore[type-arg]: StreamHandler is a generic type, but such expression would be incompatible
 # with older Python versions. Since it's not critical to mark the handler as "str only", we can
@@ -305,12 +322,11 @@ class _Formatter(logging.Formatter):
             pass
 
         # Otherwise render the message.
-        else:
-            if record.msg and record.args:
-                record.message = record.msg % record.args
+        elif record.msg and record.args:
+            record.message = record.msg % record.args
 
-            else:
-                record.message = record.msg
+        else:
+            record.message = record.msg
 
         # Original code from Formatter.format() - hard to inherit when overriding
         # Formatter.format()...
@@ -342,7 +358,8 @@ class ConsoleFormatter(_Formatter):
     def __init__(self, apply_colors: bool = True, show_timestamps: bool = False) -> None:
         super().__init__(
             '%(asctime)s %(message)s' if show_timestamps else '%(message)s',
-            apply_colors=apply_colors)
+            apply_colors=apply_colors,
+        )
 
 
 class VerbosityLevelFilter(logging.Filter):
@@ -379,7 +396,7 @@ class DebugLevelFilter(logging.Filter):
 
 class QuietnessFilter(logging.Filter):
     def filter(self, record: logging.LogRecord) -> bool:
-        if record.levelno not in (logging.DEBUG, logging.INFO):
+        if record.levelno not in (logging.DEBUG, logging.INFO, logging.WARNING):
             return True
 
         details: Optional[LogRecordDetails] = getattr(record, 'details', None)
@@ -417,13 +434,14 @@ class TopicFilter(logging.Filter):
 
 class LoggingFunction(Protocol):
     def __call__(
-            self,
-            key: str,
-            value: Optional[str] = None,
-            color: Optional[str] = None,
-            shift: int = 0,
-            level: int = 1,
-            topic: Optional[Topic] = None) -> None:
+        self,
+        key: str,
+        value: Optional[str] = None,
+        color: 'tmt.utils.themes.Style' = None,
+        shift: int = 0,
+        level: int = 1,
+        topic: Optional[Topic] = None,
+    ) -> None:
         pass
 
 
@@ -436,18 +454,18 @@ class Logger:
     """
 
     def __init__(
-            self,
-            actual_logger: logging.Logger,
-            base_shift: int = 0,
-            labels: Optional[list[str]] = None,
-            labels_padding: int = 0,
-            verbosity_level: int = DEFAULT_VERBOSITY_LEVEL,
-            debug_level: int = DEFAULT_DEBUG_LEVEL,
-            quiet: bool = False,
-            topics: Optional[set[Topic]] = None,
-            apply_colors_output: bool = True,
-            apply_colors_logging: bool = True
-            ) -> None:
+        self,
+        actual_logger: logging.Logger,
+        base_shift: int = 0,
+        labels: Optional[list[str]] = None,
+        labels_padding: int = 0,
+        verbosity_level: int = DEFAULT_VERBOSITY_LEVEL,
+        debug_level: int = DEFAULT_DEBUG_LEVEL,
+        quiet: bool = False,
+        topics: Optional[set[Topic]] = None,
+        apply_colors_output: bool = True,
+        apply_colors_logging: bool = True,
+    ) -> None:
         """
         Create a ``Logger`` instance with given verbosity levels.
 
@@ -485,24 +503,41 @@ class Logger:
         self._decolorize_output = create_decolorizer(apply_colors_output)
 
     def __repr__(self) -> str:
-        return '<Logger:' \
-            f' name={self._logger.name}' \
-            f' verbosity={self.verbosity_level}' \
-            f' debug={self.debug_level}' \
-            f' quiet={self.quiet}' \
-            f' topics={self.topics}' \
-            f' apply_colors_output={self.apply_colors_output}' \
-            f' apply_colors_logging={self.apply_colors_logging}' \
-            '>'
+        return (
+            f'<Logger:'
+            f' name={self._logger.name}'
+            f' verbosity={self.verbosity_level}'
+            f' debug={self.debug_level}'
+            f' quiet={self.quiet}'
+            f' topics={self.topics}'
+            f' apply_colors_output={self.apply_colors_output}'
+            f' apply_colors_logging={self.apply_colors_logging}'
+            f'>'
+        )
+
+    @property
+    def apply_colors_output(self) -> bool:
+        return self._apply_colors_output
+
+    @apply_colors_output.setter
+    def apply_colors_output(self, value: bool) -> None:
+        self._apply_colors_output = value
+
+        self._decolorize_output = create_decolorizer(self._apply_colors_output)
 
     @property
     def labels_span(self) -> int:
-        """ Length of rendered labels """
+        """
+        Length of rendered labels
+        """
+
         return len(render_labels(self.labels))
 
     @staticmethod
     def _normalize_logger(logger: logging.Logger) -> logging.Logger:
-        """ Reset properties of a given :py:class:`logging.Logger` instance """
+        """
+        Reset properties of a given :py:class:`logging.Logger` instance
+        """
 
         logger.propagate = True
         logger.level = logging.DEBUG
@@ -529,14 +564,14 @@ class Logger:
             quiet=self.quiet,
             topics=self.topics,
             apply_colors_output=self.apply_colors_output,
-            apply_colors_logging=self.apply_colors_logging
-            )
+            apply_colors_logging=self.apply_colors_logging,
+        )
 
     def descend(
-            self,
-            logger_name: Optional[str] = None,
-            extra_shift: int = 1
-            ) -> 'Logger':
+        self,
+        logger_name: Optional[str] = None,
+        extra_shift: int = 1,
+    ) -> 'Logger':
         """
         Create a copy of this logger instance, but with a new raw logger.
 
@@ -563,11 +598,13 @@ class Logger:
             quiet=self.quiet,
             topics=self.topics,
             apply_colors_output=self.apply_colors_output,
-            apply_colors_logging=self.apply_colors_logging
-            )
+            apply_colors_logging=self.apply_colors_logging,
+        )
 
     def add_logfile_handler(self, filepath: 'tmt.utils.Path') -> None:
-        """ Attach a log file handler to this logger """
+        """
+        Attach a log file handler to this logger
+        """
 
         handler = LogfileHandler(filepath)
 
@@ -587,9 +624,11 @@ class Logger:
 
         handler = ConsoleHandler(stream=sys.stderr)
 
-        handler.setFormatter(ConsoleFormatter(
-            apply_colors=self.apply_colors_logging,
-            show_timestamps=show_timestamps))
+        handler.setFormatter(
+            ConsoleFormatter(
+                apply_colors=self.apply_colors_logging, show_timestamps=show_timestamps
+            )
+        )
 
         handler.addFilter(VerbosityLevelFilter())
         handler.addFilter(DebugLevelFilter())
@@ -599,9 +638,10 @@ class Logger:
         self._logger.addHandler(handler)
 
     def apply_verbosity_options(
-            self,
-            cli_invocation: Optional['tmt.cli.CliInvocation'] = None,
-            **kwargs: Any) -> 'Logger':
+        self,
+        cli_invocation: Optional['tmt.cli.CliInvocation'] = None,
+        **kwargs: Any,
+    ) -> 'Logger':
         """
         Update logger's settings to match given CLI options.
 
@@ -653,17 +693,19 @@ class Logger:
 
                 raise tmt.utils.GeneralError(
                     f'Logging topic "{topic_spec}" is invalid.'
-                    f" Possible choices are {', '.join(topic.value for topic in Topic)}")
+                    f" Possible choices are {', '.join(topic.value for topic in Topic)}"
+                )
 
         return self
 
     @classmethod
     def create(
-            cls,
-            actual_logger: Optional[logging.Logger] = None,
-            apply_colors_output: bool = True,
-            apply_colors_logging: bool = True,
-            **verbosity_options: Any) -> 'Logger':
+        cls,
+        actual_logger: Optional[logging.Logger] = None,
+        apply_colors_output: bool = True,
+        apply_colors_logging: bool = True,
+        **verbosity_options: Any,
+    ) -> 'Logger':
         """
         Create a (root) tmt logger.
 
@@ -683,15 +725,15 @@ class Logger:
         return Logger(
             actual_logger,
             apply_colors_output=apply_colors_output,
-            apply_colors_logging=apply_colors_logging) \
-            .apply_verbosity_options(**verbosity_options)
+            apply_colors_logging=apply_colors_logging,
+        ).apply_verbosity_options(**verbosity_options)
 
     def _log(
-            self,
-            level: int,
-            details: LogRecordDetails,
-            message: str = ''
-            ) -> None:
+        self,
+        level: int,
+        details: LogRecordDetails,
+        message: str = '',
+    ) -> None:
         """
         Emit a log record describing the message and related properties.
 
@@ -718,54 +760,61 @@ class Logger:
                 color=details.color,
                 level=details.shift,
                 labels=self.labels,
-                labels_padding=self.labels_padding)
+                labels_padding=self.labels_padding,
+            )
 
         self._logger._log(level, message, (), extra={'details': details})
 
-    def print(
-            self,
-            text: str,
-            color: Optional[str] = None,
-            shift: int = 0,
-            ) -> None:
+    def print_format(
+        self,
+        text: str,
+        color: 'tmt.utils.themes.Style' = None,
+        shift: int = 0,
+    ) -> str:
+        """
+        Format the given text in a way suitable for :py:meth:`print`
+        """
 
-        message = indent(
+        text = indent(
             text,
             # Always apply colors - message can be decolorized later.
             color=color,
             level=shift + self._base_shift,
             labels=self.labels,
-            labels_padding=self.labels_padding)
+            labels_padding=self.labels_padding,
+        )
 
-        message = self._decolorize_output(message)
+        return self._decolorize_output(text)
 
-        print(message)
+    def print(
+        self,
+        text: str,
+        color: 'tmt.utils.themes.Style' = None,
+        shift: int = 0,
+        file: Optional[TextIO] = None,
+    ) -> None:
+        file = file or sys.stdout
+
+        print(self.print_format(text, color=color, shift=shift), file=file)
 
     def info(
-            self,
-            key: str,
-            value: Optional[LoggableValue] = None,
-            color: Optional[str] = None,
-            shift: int = 0
-            ) -> None:
-        self._log(
-            logging.INFO,
-            LogRecordDetails(
-                key=key,
-                value=value,
-                color=color,
-                shift=shift)
-            )
+        self,
+        key: str,
+        value: Optional[LoggableValue] = None,
+        color: 'tmt.utils.themes.Style' = None,
+        shift: int = 0,
+    ) -> None:
+        self._log(logging.INFO, LogRecordDetails(key=key, value=value, color=color, shift=shift))
 
     def verbose(
-            self,
-            key: str,
-            value: Optional[LoggableValue] = None,
-            color: Optional[str] = None,
-            shift: int = 0,
-            level: int = 1,
-            topic: Optional[Topic] = None
-            ) -> None:
+        self,
+        key: str,
+        value: Optional[LoggableValue] = None,
+        color: 'tmt.utils.themes.Style' = None,
+        shift: int = 0,
+        level: int = 1,
+        topic: Optional[Topic] = None,
+    ) -> None:
         self._log(
             logging.INFO,
             LogRecordDetails(
@@ -774,18 +823,19 @@ class Logger:
                 color=color,
                 shift=shift,
                 message_verbosity_level=level,
-                message_topic=topic)
-            )
+                message_topic=topic,
+            ),
+        )
 
     def debug(
-            self,
-            key: str,
-            value: Optional[LoggableValue] = None,
-            color: Optional[str] = None,
-            shift: int = 0,
-            level: int = 1,
-            topic: Optional[Topic] = None
-            ) -> None:
+        self,
+        key: str,
+        value: Optional[LoggableValue] = None,
+        color: 'tmt.utils.themes.Style' = None,
+        shift: int = 0,
+        level: int = 1,
+        topic: Optional[Topic] = None,
+    ) -> None:
         self._log(
             logging.DEBUG,
             LogRecordDetails(
@@ -794,36 +844,36 @@ class Logger:
                 color=color,
                 shift=shift,
                 message_debug_level=level,
-                message_topic=topic)
-            )
+                message_topic=topic,
+            ),
+        )
 
-    def warn(
-            self,
-            message: str,
-            shift: int = 0
-            ) -> None:
+    def warning(
+        self,
+        message: str,
+        shift: int = 0,
+    ) -> None:
         self._log(
-            logging.WARN,
-            LogRecordDetails(
-                key='warn',
-                value=message,
-                color='yellow',
-                shift=shift)
-            )
+            logging.WARNING,
+            LogRecordDetails(key='warn', value=message, color='yellow', shift=shift),
+        )
+
+    @deprecated("Use Logger.warning instead")
+    def warn(
+        self,
+        message: str,
+        shift: int,
+    ) -> None:
+        return self.warning(message, shift)
 
     def fail(
-            self,
-            message: str,
-            shift: int = 0
-            ) -> None:
+        self,
+        message: str,
+        shift: int = 0,
+    ) -> None:
         self._log(
-            logging.ERROR,
-            LogRecordDetails(
-                key='fail',
-                value=message,
-                color='red',
-                shift=shift)
-            )
+            logging.ERROR, LogRecordDetails(key='fail', value=message, color='red', shift=shift)
+        )
 
     _bootstrap_logger: Optional['Logger'] = None
 
@@ -836,11 +886,12 @@ class Logger:
 
             This logger has a **very** limited use case span, i.e.
             before tmt can digest its command-line options and create a
-            proper logger. This happens inside :py:func:`tmt.cli.main`
-            function, but there are some actions taken by tmt code
-            before this function is called by Click, actions that need
-            to emit logging messages. Using it anywhere outside of this
-            brief time in tmt's runtime should be ruled out.
+            proper logger. This happens inside
+            :py:func:`tmt.cli._root.main` function, but there are some
+            actions taken by tmt code before this function is called by
+            Click, actions that need to emit logging messages. Using it
+            anywhere outside of this brief time in tmt's runtime should
+            be ruled out.
         """
 
         if cls._bootstrap_logger is None:
