@@ -1,7 +1,9 @@
-""" Common options and the MethodCommand class """
+"""
+Common options and the MethodCommand class
+"""
 
 import contextlib
-import dataclasses
+import pathlib
 import re
 import textwrap
 from collections.abc import Sequence
@@ -12,6 +14,7 @@ import click
 import tmt.lint
 import tmt.log
 import tmt.utils
+from tmt.container import container
 
 # When dealing with older Click packages (I'm looking at you, Python 3.6),
 # we need to define FC on our own.
@@ -29,9 +32,11 @@ if TYPE_CHECKING:
     import tmt.utils
 
 
-@dataclasses.dataclass(frozen=True)
+@container(frozen=True)
 class Deprecated:
-    """ Version information and hint for obsolete options """
+    """
+    Version information and hint for obsolete options
+    """
 
     since: str
     hint: Optional[str] = None
@@ -46,6 +51,63 @@ class Deprecated:
         return f'{message}.'
 
 
+class Path(click.ParamType):
+    name = 'path'
+
+    def convert(
+        self,
+        value: Any,
+        param: Optional[click.Parameter],
+        ctx: Optional[click.Context],
+    ) -> Optional[tmt.utils.Path]:
+        """
+        Convert the value to the correct type. This is not called if
+        the value is ``None`` (the missing value).
+
+        This must accept string values from the command line, as well as
+        values that are already the correct type. It may also convert
+        other compatible types.
+
+        The ``param`` and ``ctx`` arguments may be ``None`` in certain
+        situations, such as when converting prompt input.
+
+        If the value cannot be converted, call :meth:`fail` with a
+        descriptive message.
+
+        :param value: The value to convert.
+        :param param: The parameter that is using this type to convert
+            its value. May be ``None``.
+        :param ctx: The current context that arrived at this value. May
+            be ``None``.
+        """
+
+        if value is None:
+            return None
+
+        if isinstance(value, str):
+            return tmt.utils.Path(value)
+
+        if isinstance(value, tmt.utils.Path):
+            return value
+
+        if isinstance(value, pathlib.Path):  # noqa: TID251
+            return tmt.utils.Path(str(value))
+
+        if param is None:
+            self.fail(
+                f"A path-like string was expected, '{type(value).__name__}' found.",
+                param=param,
+                ctx=ctx,
+            )
+
+        # RET503: ruff does not recognize NoReturn annotation of `self.fail`.
+        self.fail(  # noqa: RET503
+            f"Field '{param.name}' must be a path-like string, '{type(value).__name__}' found.",
+            param=param,
+            ctx=ctx,
+        )
+
+
 MethodDictType = dict[str, click.core.Command]
 
 # Originating in click.decorators, an opaque type describing "decorator" functions
@@ -63,25 +125,27 @@ ClickOptionDecoratorType = _ClickOptionDecoratorType[Any]
 
 
 def option(
-        *param_decls: str,
-        # Following parameters are inherited from click.option()/Option/Parameter.
-        # May allow stricter types than the original, because tmt code base does not
-        # care for every Click use case.
-        show_default: bool = False,
-        is_flag: bool = False,
-        multiple: bool = False,
-        count: bool = False,
-        type: Optional[Union[click.Choice, Any]] = None,
-        help: Optional[str] = None,
-        required: bool = False,
-        default: Optional[Any] = None,
-        nargs: Optional[int] = None,
-        metavar: Optional[str] = None,
-        prompt: Optional[str] = None,
-        envvar: Optional[str] = None,
-        # Following parameters are our additions.
-        choices: Optional[Sequence[str]] = None,
-        deprecated: Optional[Deprecated] = None) -> ClickOptionDecoratorType:
+    *param_decls: str,
+    # Following parameters are inherited from click.option()/Option/Parameter.
+    # May allow stricter types than the original, because tmt code base does not
+    # care for every Click use case.
+    show_default: bool = False,
+    is_flag: bool = False,
+    multiple: bool = False,
+    count: bool = False,
+    type: Optional[Union[click.Choice, Any]] = None,  # noqa: A002  `type` is shadowing a Python builtin
+    help: Optional[str] = None,
+    required: bool = False,
+    default: Optional[Any] = None,
+    nargs: Optional[int] = None,
+    metavar: Optional[str] = None,
+    prompt: Optional[str] = None,
+    envvar: Optional[str] = None,
+    hidden: bool = False,
+    # Following parameters are our additions.
+    choices: Optional[Sequence[str]] = None,
+    deprecated: Optional[Deprecated] = None,
+) -> ClickOptionDecoratorType:
     """
     Attaches an option to the command.
 
@@ -108,7 +172,7 @@ def option(
             help = deprecated.rendered
 
     if choices is not None:
-        type = click.Choice(choices)
+        type = click.Choice(choices)  # noqa: A001  `type` is shadowing a Python builtin
 
     # Add a metavar listing choices unless an explicit metavar has been provided
     if isinstance(type, click.Choice) and metavar is None:
@@ -128,161 +192,268 @@ def option(
 # Verbose, debug and quiet output
 VERBOSITY_OPTIONS: list[ClickOptionDecoratorType] = [
     option(
-        '-v', '--verbose', count=True, default=0,
-        help='Show more details. Use multiple times to raise verbosity.'),
+        '-v',
+        '--verbose',
+        count=True,
+        default=0,
+        help='Show more details. Use multiple times to raise verbosity.',
+    ),
     option(
-        '-d', '--debug', count=True, default=0,
-        help='Provide debugging information. Repeat to see more details.'),
+        '-d',
+        '--debug',
+        count=True,
+        default=0,
+        help='Provide debugging information. Repeat to see more details.',
+    ),
     option(
-        '-q', '--quiet', is_flag=True,
-        help='Be quiet. Exit code is just enough for me.'),
+        '-q',
+        '--quiet',
+        is_flag=True,
+        help='Be quiet. Exit code is just enough for me.',
+    ),
     option(
         '--log-topic',
         choices=[topic.value for topic in tmt.log.Topic],
         multiple=True,
-        help='If specified, --debug and --verbose would emit logs also for these topics.')
-    ]
+        help='If specified, --debug and --verbose would emit logs also for these topics.',
+    ),
+]
 
-# Force, dry and run again actions
+# Force, dry, feeling-safe and run again actions
 DRY_OPTIONS: list[ClickOptionDecoratorType] = [
     option(
-        '-n', '--dry', is_flag=True, default=False,
-        help='Run in dry mode. No changes, please.'),
-    ]
+        '-n',
+        '--dry',
+        is_flag=True,
+        default=False,
+        help='Run in dry mode. No changes, please.',
+    ),
+]
 
 FORCE_DRY_OPTIONS: list[ClickOptionDecoratorType] = [
     option(
-        '-f', '--force', is_flag=True,
-        help='Overwrite existing files and step data.'),
-    *DRY_OPTIONS]
+        '-f',
+        '--force',
+        is_flag=True,
+        help='Overwrite existing files and step data.',
+    ),
+    *DRY_OPTIONS,
+]
 
 AGAIN_OPTION: list[ClickOptionDecoratorType] = [
     option(
-        '--again', is_flag=True,
-        help='Run again, even if already done before.'),
-    ]
+        '--again',
+        is_flag=True,
+        help='Run again, even if already done before.',
+    ),
+]
+
+FEELING_SAFE_OPTION: list[ClickOptionDecoratorType] = [
+    option(
+        '--feeling-safe',
+        metavar='FEELING_SAFE',
+        envvar='TMT_FEELING_SAFE',
+        is_flag=True,
+        default=False,
+        help="""
+             WARNING: with this option, tmt would be allowed to make
+             potentially dangerous actions. For example, some metadata
+             keys may cause scripts being executed on the runner.
+             Do not use this option unless you trust metadata consumed
+             by tmt, or unless you know what you are doing.
+             """,
+    )
+]
 
 # Fix action
 FIX_OPTIONS: list[ClickOptionDecoratorType] = [
     option('-F', '--fix', is_flag=True, help='Attempt to fix all discovered issues.')
-    ]
+]
 
 WORKDIR_ROOT_OPTIONS: list[ClickOptionDecoratorType] = [
     option(
-        '--workdir-root', metavar='PATH', envvar='TMT_WORKDIR_ROOT',
+        '--workdir-root',
+        metavar='PATH',
+        envvar='TMT_WORKDIR_ROOT',
         default=tmt.utils.WORKDIR_ROOT,
+        type=Path(),
         help=f"""
              Path to root directory containing run workdirs.
              Defaults to '{tmt.utils.WORKDIR_ROOT}'.
-             """)
-    ]
+             """,
+    )
+]
 
 FILTER_OPTION: list[ClickOptionDecoratorType] = [
     option(
-        '-f', '--filter', 'filters', metavar='FILTER', multiple=True,
+        '-f',
+        '--filter',
+        'filters',
+        metavar='FILTER',
+        multiple=True,
         help="""
         Apply an advanced filter using key:value pairs and logical operators.
         For example 'tier:1 & tag:core'. Use the 'name' key to search by name.
         See 'pydoc fmf.filter' for detailed documentation on the syntax.
-        """),
-    ]
+        """,
+    ),
+]
 
 FILTER_OPTION_LONG: list[ClickOptionDecoratorType] = [
     option(
-        '--filter', 'filters', metavar='FILTER', multiple=True,
+        '--filter',
+        'filters',
+        metavar='FILTER',
+        multiple=True,
         help="""
         Apply an advanced filter using key:value pairs and logical operators.
         For example 'tier:1 & tag:core'. Use the 'name' key to search by name.
         See 'pydoc fmf.filter' for detailed documentation on the syntax.
-        """),
-    ]
+        """,
+    ),
+]
 
 FILTERING_OPTIONS: list[ClickOptionDecoratorType] = [
-    click.argument(
-        'names', nargs=-1, metavar='[REGEXP|.]'),
+    click.argument('names', nargs=-1, metavar='[REGEXP|.]'),
     *FILTER_OPTION,
     option(
-        '-c', '--condition', 'conditions', metavar="EXPR", multiple=True,
-        help="Use arbitrary Python expression for filtering."),
+        '-c',
+        '--condition',
+        'conditions',
+        metavar="EXPR",
+        multiple=True,
+        help="Use arbitrary Python expression for filtering.",
+    ),
     option(
-        '--enabled', is_flag=True,
-        help="Show only enabled tests, plans or stories."),
+        '--enabled',
+        is_flag=True,
+        help="Show only enabled tests, plans or stories.",
+    ),
     option(
-        '--disabled', is_flag=True,
-        help="Show only disabled tests, plans or stories."),
+        '--disabled',
+        is_flag=True,
+        help="Show only disabled tests, plans or stories.",
+    ),
     option(
-        '--link', 'links', metavar="RELATION:TARGET", multiple=True,
+        '--link',
+        'links',
+        metavar="RELATION:TARGET",
+        multiple=True,
         help="""
              Filter by linked objects (regular expressions are supported for both relation and
              target).
-             """),
+             """,
+    ),
     option(
-        '-x', '--exclude', 'exclude', metavar='[REGEXP]', multiple=True,
-        help="Exclude a regular expression from search result."),
-    ]
+        '-x',
+        '--exclude',
+        'exclude',
+        metavar='[REGEXP]',
+        multiple=True,
+        help="Exclude a regular expression from search result.",
+    ),
+]
 
 
 FILTERING_OPTIONS_LONG: list[ClickOptionDecoratorType] = [
-    click.argument(
-        'names', nargs=-1, metavar='[REGEXP|.]'),
+    click.argument('names', nargs=-1, metavar='[REGEXP|.]'),
     *FILTER_OPTION_LONG,
     option(
-        '--condition', 'conditions', metavar="EXPR", multiple=True,
-        help="Use arbitrary Python expression for filtering."),
+        '--condition',
+        'conditions',
+        metavar="EXPR",
+        multiple=True,
+        help="Use arbitrary Python expression for filtering.",
+    ),
     option(
-        '--enabled', is_flag=True,
-        help="Show only enabled tests, plans or stories."),
+        '--enabled',
+        is_flag=True,
+        help="Show only enabled tests, plans or stories.",
+    ),
     option(
-        '--disabled', is_flag=True,
-        help="Show only disabled tests, plans or stories."),
+        '--disabled',
+        is_flag=True,
+        help="Show only disabled tests, plans or stories.",
+    ),
     option(
-        '--link', 'links', metavar="RELATION:TARGET", multiple=True,
+        '--link',
+        'links',
+        metavar="RELATION:TARGET",
+        multiple=True,
         help="""
              Filter by linked objects (regular expressions are supported for both relation and
              target).
-             """),
+             """,
+    ),
     option(
-        '--exclude', 'exclude', metavar='[REGEXP]', multiple=True,
-        help="Exclude a regular expression from search result."),
-    ]
+        '--exclude',
+        'exclude',
+        metavar='[REGEXP]',
+        multiple=True,
+        help="Exclude a regular expression from search result.",
+    ),
+]
 
 
 STORY_FLAGS_FILTER_OPTIONS: list[ClickOptionDecoratorType] = [
     option(
-        '--implemented', is_flag=True,
-        help='Implemented stories only.'),
+        '--implemented',
+        is_flag=True,
+        help='Implemented stories only.',
+    ),
     option(
-        '--unimplemented', is_flag=True,
-        help='Unimplemented stories only.'),
+        '--unimplemented',
+        is_flag=True,
+        help='Unimplemented stories only.',
+    ),
     option(
-        '--verified', is_flag=True,
-        help='Stories verified by tests.'),
+        '--verified',
+        is_flag=True,
+        help='Stories verified by tests.',
+    ),
     option(
-        '--unverified', is_flag=True,
-        help='Stories not verified by tests.'),
+        '--unverified',
+        is_flag=True,
+        help='Stories not verified by tests.',
+    ),
     option(
-        '--documented', is_flag=True,
-        help='Documented stories only.'),
+        '--documented',
+        is_flag=True,
+        help='Documented stories only.',
+    ),
     option(
-        '--undocumented', is_flag=True,
-        help='Undocumented stories only.'),
+        '--undocumented',
+        is_flag=True,
+        help='Undocumented stories only.',
+    ),
     option(
-        '--covered', is_flag=True,
-        help='Covered stories only.'),
+        '--covered',
+        is_flag=True,
+        help='Covered stories only.',
+    ),
     option(
-        '--uncovered', is_flag=True,
-        help='Uncovered stories only.'),
-    ]
+        '--uncovered',
+        is_flag=True,
+        help='Uncovered stories only.',
+    ),
+]
 
 FMF_SOURCE_OPTIONS: list[ClickOptionDecoratorType] = [
     option(
-        '--source', is_flag=True, help="Select by fmf source file names instead of object names."
-        )
-    ]
+        '--source',
+        is_flag=True,
+        help="Select by fmf source file names instead of object names.",
+    )
+]
 
 REMOTE_PLAN_OPTIONS: list[ClickOptionDecoratorType] = [
-    option('-s', '--shallow', is_flag=True, help='Do not clone remote plan.')
-    ]
+    option(
+        '-s',
+        '--shallow',
+        is_flag=True,
+        help='Do not clone remote plan.',
+    )
+]
 
 
 _lint_outcomes = [member.value for member in tmt.lint.LinterOutcome.__members__.values()]
@@ -291,49 +462,57 @@ LINT_OPTIONS: list[ClickOptionDecoratorType] = [
     option(
         '--list-checks',
         is_flag=True,
-        help='List all available checks.'),
+        help='List all available checks.',
+    ),
     option(
         '--enable-check',
         'enable_checks',
         metavar='CHECK-ID',
         multiple=True,
         type=str,
-        help='Run only checks mentioned by this option.'),
+        help='Run only checks mentioned by this option.',
+    ),
     option(
         '--disable-check',
         'disable_checks',
         metavar='CHECK-ID',
         multiple=True,
         type=str,
-        help='Do not run checks mentioned by this option.'),
+        help='Do not run checks mentioned by this option.',
+    ),
     option(
         '--enforce-check',
         'enforce_checks',
         metavar='CHECK-ID',
         multiple=True,
         type=str,
-        help='Consider linting as failed if any of the mentioned checks is not a pass.'),
+        help='Consider linting as failed if any of the mentioned checks is not a pass.',
+    ),
     option(
         '--failed-only',
         is_flag=True,
-        help='Display only tests/plans/stories that fail a check.'),
+        help='Display only tests/plans/stories that fail a check.',
+    ),
     option(
         '--outcome-only',
         multiple=True,
         choices=_lint_outcomes,
-        help='Display only checks with the given outcome.')
-    ]
+        help='Display only checks with the given outcome.',
+    ),
+]
 
 
 ENVIRONMENT_OPTIONS: list[ClickOptionDecoratorType] = [
     option(
-        '-e', '--environment',
+        '-e',
+        '--environment',
         metavar='KEY=VALUE|@FILE',
         multiple=True,
         help="""
              Set environment variable. Can be specified multiple times. The "@" prefix marks a file
              to load (yaml or dotenv formats supported).
-             """),
+             """,
+    ),
     option(
         '--environment-file',
         metavar='FILE|URL',
@@ -341,8 +520,9 @@ ENVIRONMENT_OPTIONS: list[ClickOptionDecoratorType] = [
         help="""
              Set environment variables from file or url (yaml or dotenv formats are supported). Can
              be specified multiple times.
-             """)
-    ]
+             """,
+    ),
+]
 
 
 def create_options_decorator(options: list[ClickOptionDecoratorType]) -> Callable[[FC], FC]:
@@ -355,51 +535,6 @@ def create_options_decorator(options: list[ClickOptionDecoratorType]) -> Callabl
     return common_decorator
 
 
-def show_step_method_hints(
-        step_name: str,
-        how: str,
-        logger: tmt.log.Logger) -> None:
-    """
-    Show hints about available step methods' installation
-
-    The logger will be used to output the hints to the terminal, hence
-    it must be an instance of a subclass of tmt.utils.Common (info method
-    must be available).
-    """
-    if step_name == 'provision':
-        if how == 'virtual':
-            logger.info(
-                'hint', "Install 'tmt+provision-virtual' "
-                        "to run tests in a virtual machine.", color='blue')
-        if how == 'container':
-            logger.info(
-                'hint', "Install 'tmt+provision-container' "
-                        "to run tests in a container.", color='blue')
-        if how == 'minute':
-            logger.info(
-                'hint', "Install 'tmt-redhat-provision-minute' "
-                        "to run tests in 1minutetip OpenStack backend. "
-                        "(Available only from the internal COPR repository.)",
-                        color='blue')
-        logger.info(
-            'hint', "Use the 'local' method to execute tests "
-                    "directly on your localhost.", color='blue')
-        logger.info(
-            'hint', "See 'tmt run provision --help' for all "
-                    "available provision options.", color='blue')
-    elif step_name == 'report':
-        if how == 'junit':
-            logger.info(
-                'hint', "Install 'tmt+report-junit' to write results "
-                        "in JUnit format.", color='blue')
-        logger.info(
-            'hint', "Use the 'display' method to show test results "
-                    "on the terminal.", color='blue')
-        logger.info(
-            'hint', "See 'tmt run report --help' for all "
-                    "available report options.", color='blue')
-
-
 def create_method_class(methods: MethodDictType) -> type[click.Command]:
     """
     Create special class to handle different options for each method
@@ -410,14 +545,20 @@ def create_method_class(methods: MethodDictType) -> type[click.Command]:
     """
 
     def is_likely_subcommand(arg: str, subcommands: list[str]) -> bool:
-        """ Return true if arg is the beginning characters of a subcommand """
+        """
+        Return true if arg is the beginning characters of a subcommand
+        """
+
         return any(subcommand.startswith(arg) for subcommand in subcommands)
 
     class MethodCommand(click.Command):
         _method: Optional[click.Command] = None
 
         def _check_method(self, context: 'tmt.cli.Context', args: list[str]) -> None:
-            """ Manually parse the --how option """
+            """
+            Manually parse the --how option
+            """
+
             # Avoiding circular imports
             import tmt.steps
 
@@ -428,8 +569,7 @@ def create_method_class(methods: MethodDictType) -> type[click.Command]:
             import tmt.utils
 
             how = None
-            subcommands = (
-                tmt.steps.STEPS + tmt.steps.ACTIONS + ['tests', 'plans'])
+            subcommands = tmt.steps.STEPS + tmt.steps.ACTIONS + ['tests', 'plans']
 
             def _find_option_by_arg(arg: str) -> Optional[click.Parameter]:
                 for option in self.params:
@@ -507,7 +647,8 @@ def create_method_class(methods: MethodDictType) -> type[click.Command]:
                     # been claimed by an option. It is highly likely this is a misspelled
                     # subcommand name.
                     raise tmt.utils.SpecificationError(
-                        f"Invalid subcommand of 'run' found: '{arg}'.")
+                        f"Invalid subcommand of 'run' found: '{arg}'."
+                    )
 
                 return None
 
@@ -522,18 +663,22 @@ def create_method_class(methods: MethodDictType) -> type[click.Command]:
                         break
 
             if how and self._method is None:
+                from tmt.utils.hints import print_hint
+
                 # Use run for logging, steps may not be initialized yet
                 assert context.obj.run is not None  # narrow type
                 assert self.name is not None  # narrow type
-                show_step_method_hints(self.name, how, context.obj.run._logger)
-                raise tmt.utils.SpecificationError(
-                    f"Unsupported {self.name} method '{how}'.")
+
+                print_hint(
+                    id_=f'{self.name}/{how}', ignore_missing=True, logger=context.obj.run._logger
+                )
+                print_hint(id_=self.name, ignore_missing=True, logger=context.obj.run._logger)
+
+                raise tmt.utils.SpecificationError(f"Unsupported {self.name} method '{how}'.")
 
         def parse_args(  # type: ignore[override]
-                self,
-                context: 'tmt.cli.Context',
-                args: list[str]
-                ) -> list[str]:
+            self, context: 'tmt.cli.Context', args: list[str]
+        ) -> list[str]:
             self._check_method(context, args)
             if self._method is not None:
                 return self._method.parse_args(context, args)
