@@ -1,5 +1,4 @@
 import contextlib
-import dataclasses
 import glob
 import os
 import re
@@ -17,14 +16,18 @@ import tmt.options
 import tmt.steps
 import tmt.steps.discover
 import tmt.utils
+import tmt.utils.git
+from tmt.base import _RawAdjustRule
+from tmt.container import container, field
 from tmt.steps.prepare.distgit import insert_to_prepare_step
-from tmt.utils import Command, Environment, EnvVarValue, Path, field
+from tmt.utils import Command, Environment, EnvVarValue, Path
 
 
 def normalize_ref(
-        key_address: str,
-        value: Optional[Any],
-        logger: tmt.log.Logger) -> Optional[str]:
+    key_address: str,
+    value: Optional[Any],
+    logger: tmt.log.Logger,
+) -> Optional[str]:
     if value is None:
         return None
 
@@ -34,25 +37,51 @@ def normalize_ref(
     raise tmt.utils.NormalizationError(key_address, value, 'unset or a string')
 
 
-@dataclasses.dataclass
+@container
 class DiscoverFmfStepData(tmt.steps.discover.DiscoverStepData):
     # Basic options
     url: Optional[str] = field(
         default=cast(Optional[str], None),
         option=('-u', '--url'),
         metavar='REPOSITORY',
-        help='URL of the git repository with fmf metadata.')
+        help="""
+            Git repository containing the metadata tree.
+            Current git repository used by default.
+            """,
+    )
+
     ref: Optional[str] = field(
         default=cast(Optional[str], None),
         option=('-r', '--ref'),
         metavar='REVISION',
-        help='Branch, tag or commit specifying the git revision.',
-        normalize=normalize_ref)
+        help="""
+            Branch, tag or commit specifying the desired git
+            revision. Defaults to the remote repository's default
+            branch if ``url`` was set or to the current ``HEAD``
+            of the current repository.
+
+            Additionally, one can set ``ref`` dynamically.
+            This is possible using a special file in tmt format
+            stored in the *default* branch of a tests repository.
+            This special file should contain rules assigning attribute ``ref``
+            in an *adjust* block, for example depending on a test run context.
+
+            Dynamic ``ref`` assignment is enabled whenever a test plan
+            reference has the format ``ref: @FILEPATH``.
+            """,
+        normalize=normalize_ref,
+    )
+
     path: Optional[str] = field(
         default=cast(Optional[str], None),
         option=('-p', '--path'),
         metavar='ROOT',
-        help='Path to the metadata tree root.')
+        help="""
+            Path to the metadata tree root. Must be relative to
+            the git repository root if ``url`` was provided, absolute
+            local filesystem path otherwise. By default ``.`` is used.
+            """,
+    )
 
     # Selecting tests
     test: list[str] = field(
@@ -60,51 +89,83 @@ class DiscoverFmfStepData(tmt.steps.discover.DiscoverStepData):
         option=('-t', '--test'),
         metavar='NAMES',
         multiple=True,
-        help='Select tests by name.',
-        normalize=tmt.utils.normalize_string_list)
+        help="""
+            List of test names or regular expressions used to
+            select tests by name. Duplicate test names are allowed
+            to enable repetitive test execution, preserving the
+            listed test order.
+            """,
+        normalize=tmt.utils.normalize_string_list,
+    )
+
     link: list[str] = field(
         default_factory=list,
         option='--link',
         metavar="RELATION:TARGET",
         multiple=True,
         help="""
-             Filter by linked objects (regular expressions are supported for both relation and
-             target).
-             """)
+            Select tests using the :ref:`/spec/core/link` keys.
+            Values must be in the form of ``RELATION:TARGET``,
+            tests containing at least one of them are selected.
+            Regular expressions are supported for both relation
+            and target. Relation part can be omitted to match all
+            relations.
+             """,
+    )
+
     filter: list[str] = field(
         default_factory=list,
         option=('-F', '--filter'),
         metavar='FILTERS',
         multiple=True,
-        help='Include only tests matching the filter.',
-        normalize=tmt.utils.normalize_string_list)
+        help="""
+            Apply advanced filter based on test metadata attributes.
+            See ``pydoc fmf.filter`` for more info.
+            """,
+        normalize=tmt.utils.normalize_string_list,
+    )
     exclude: list[str] = field(
         default_factory=list,
         option=('-x', '--exclude'),
         metavar='REGEXP',
         multiple=True,
         help="Exclude tests matching given regular expression.",
-        normalize=tmt.utils.normalize_string_list)
+        normalize=tmt.utils.normalize_string_list,
+    )
 
     # Modified only
     modified_only: bool = field(
         default=False,
         option=('-m', '--modified-only'),
         is_flag=True,
-        help='If set, select only tests modified since reference revision.')
+        help="""
+            Set to ``true`` if you want to filter modified tests
+            only. The test is modified if its name starts with
+            the name of any directory modified since ``modified-ref``.
+            """,
+    )
+
     modified_url: Optional[str] = field(
         default=cast(Optional[str], None),
         option='--modified-url',
         metavar='REPOSITORY',
-        help='URL of the reference git repository with fmf metadata.')
+        help="""
+            An additional remote repository to be used as the
+            reference for comparison. Will be fetched as a
+            reference remote in the test dir.
+            """,
+    )
+
     modified_ref: Optional[str] = field(
         default=cast(Optional[str], None),
         option='--modified-ref',
         metavar='REVISION',
         help="""
-            Branch, tag or commit specifying the reference git revision (if not provided, the
-            default branch is used).
-            """)
+            The branch, tag or commit specifying the reference git revision (if not provided, the
+            default branch is used). Note that you need to specify ``reference/<branch>`` to
+            compare to a branch from the repository specified in ``modified-url``.
+            """,
+    )
 
     # Dist git integration
     dist_git_init: bool = field(
@@ -112,8 +173,12 @@ class DiscoverFmfStepData(tmt.steps.discover.DiscoverStepData):
         option='--dist-git-init',
         is_flag=True,
         help="""
-             Initialize fmf root inside extracted sources (at dist-git-extract or top directory).
-             """)
+             Set to ``true`` to initialize fmf root inside extracted sources at
+             ``dist-git-extract`` location or top directory. To be used when the
+             sources contain fmf files (for example tests) but do not have an
+             associated fmf root.
+             """,
+    )
     dist_git_remove_fmf_root: bool = field(
         default=False,
         option='--dist-git-remove-fmf-root',
@@ -121,19 +186,27 @@ class DiscoverFmfStepData(tmt.steps.discover.DiscoverStepData):
         help="""
              Remove fmf root from extracted source (top one or selected by copy-path, happens
              before dist-git-extract.
-             """)
+             """,
+    )
     dist_git_merge: bool = field(
         default=False,
         option='--dist-git-merge',
         is_flag=True,
-        help='Merge copied sources and plan fmf root.')
+        help="""
+            Set to ``true`` to combine fmf root from the sources and fmf root from the plan.
+            It allows to have plans and tests defined in the DistGit repo which use tests
+            and other resources from the downloaded sources. Any plans in extracted sources
+            will not be processed.
+            """,
+    )
     dist_git_extract: Optional[str] = field(
         default=cast(Optional[str], None),
         option='--dist-git-extract',
         help="""
              What to copy from extracted sources, globbing is supported. Defaults to the top fmf
              root if it is present, otherwise top directory (shortcut "/").
-             """)
+             """,
+    )
 
     # Special options
     sync_repo: bool = field(
@@ -143,18 +216,31 @@ class DiscoverFmfStepData(tmt.steps.discover.DiscoverStepData):
         help="""
              Force the sync of the whole git repo. By default, the repo is copied only if the used
              options require it.
-             """)
+             """,
+    )
     fmf_id: bool = field(
         default=False,
         option='--fmf-id',
         is_flag=True,
-        help='Only print fmf identifiers of discovered tests to the standard output and exit.')
+        help='Only print fmf identifiers of discovered tests to the standard output and exit.',
+    )
     prune: bool = field(
         default=False,
         option=('--prune / --no-prune'),
         is_flag=True,
         show_default=True,
-        help="Copy only immediate directories of executed tests and their required files.")
+        help="Copy only immediate directories of executed tests and their required files.",
+    )
+
+    # Edit discovered tests
+    adjust_tests: Optional[list[_RawAdjustRule]] = field(
+        default_factory=list,
+        normalize=tmt.utils.normalize_adjust,
+        help="""
+             Modify metadata of discovered tests from the plan itself. Use the
+             same format as for adjust rules.
+             """,
+    )
 
     # Upgrade plan path so the plan is not pruned
     upgrade_path: Optional[str] = None
@@ -164,9 +250,10 @@ class DiscoverFmfStepData(tmt.steps.discover.DiscoverStepData):
     revision: Optional[str] = None
 
     def post_normalization(
-            self,
-            raw_data: tmt.steps._RawStepData,
-            logger: tmt.log.Logger) -> None:
+        self,
+        raw_data: tmt.steps._RawStepData,
+        logger: tmt.log.Logger,
+    ) -> None:
         super().post_normalization(raw_data, logger)
 
         if self.repository:
@@ -203,8 +290,26 @@ class DiscoverFmf(tmt.steps.discover.DiscoverPlugin[DiscoverFmfStepData]):
 
     If no ``ref`` is provided, the default branch from the origin is used.
 
-    For DistGit repo one can extract sources first and discover
-    tests from it by using ``distgit-source: true`` later in Prepare step.
+    The following keys are used to limit the test discovery:
+
+    ``test`` - list of test names or regular expressions used to select tests
+
+    ``link`` - select tests using the link keys
+
+    ``filter`` - apply advanced filter based on test metadata attributes
+
+    ``exclude`` - exclude tests which match a regular expression
+
+    ``prune`` - copy only immediate directories of executed tests and their required files
+
+
+    For DistGit repo one can download sources and use code from them in
+    the tests. Sources are extracted into ``$TMT_SOURCE_DIR`` path,
+    patches are applied by default. See options to install build
+    dependencies or to just download sources without applying patches.
+    To apply patches, special ``prepare`` phase with order ``60`` is
+    added, and ``prepare`` step has to be enabled for it to run.
+
     It can be used together with ``ref``, ``path`` and ``url``,
     however ``ref`` is not possible without using ``url``.
 
@@ -214,17 +319,6 @@ class DiscoverFmf(tmt.steps.discover.DiscoverPlugin[DiscoverFmfStepData]):
             how: fmf
             dist-git-source: true
 
-    Related config options (all optional):
-
-    * ``dist-git-merge`` - set to ``true`` if you want to copy in extracted
-      sources to the local repo
-    * ``dist-git-init`` - set to ``true`` and ``fmf init`` will be called inside
-      extracted sources (at ``dist-git-extract`` or top directory)
-    * ``dist-git-extract`` - directory (glob supported) to copy from
-      extracted sources (defaults to inner fmf root)
-    * ``dist-git-remove-fmf-root`` - set to ``true`` to remove fmf root from
-      extracted sources
-
     Selecting tests containing specified link is possible using ``link``
     key accepting ``RELATION:TARGET`` format of values. Regular
     expressions are supported for both relation and target part of the
@@ -240,11 +334,7 @@ class DiscoverFmf(tmt.steps.discover.DiscoverPlugin[DiscoverFmfStepData]):
     in git since a given revision. This can be particularly useful when
     testing changes to tests themselves (e.g. in a pull request CI).
 
-    Related config options (all optional):
-
-    * ``modified-only`` - set to ``true`` if you want to filter modified tests
-    * ``modified-url`` - fetched as "reference" remote in the test dir
-    * ``modified-ref`` - the ref to compare against
+    Related keys: ``modified-only``, ``modified-url``, ``modified-ref``
 
     Example to compare local repo against upstream ``main`` branch:
 
@@ -259,6 +349,29 @@ class DiscoverFmf(tmt.steps.discover.DiscoverPlugin[DiscoverFmfStepData]):
     Note that internally the modified tests are appended to the list
     specified via ``test``, so those tests will also be selected even if
     not modified.
+
+    Use the ``adjust-tests`` key to modify the discovered tests'
+    metadata directly from the plan. For example, extend the test
+    duration for slow hardware or modify the list of required packages
+    when you do not have write access to the remote test repository.
+    The value should follow the ``adjust`` rules syntax.
+
+    The following example adds an ``avc`` check for each discovered
+    test, doubles its duration and replaces each occurrence of the word
+    ``python3.11`` in the list of required packages.
+
+    .. code-block:: yaml
+
+        discover:
+            how: fmf
+            adjust-tests:
+              - check+:
+                  - how: avc
+              - duration+: '*2'
+                because: Slow system under test
+                when: arch == i286
+              - require~:
+                  - '/python3.11/python3.12/'
     """
 
     _data_class = DiscoverFmfStepData
@@ -269,27 +382,37 @@ class DiscoverFmf(tmt.steps.discover.DiscoverPlugin[DiscoverFmfStepData]):
         "modified-url",
         "modified-only",
         "fmf-id",
-        )
+    )
 
     @property
     def is_in_standalone_mode(self) -> bool:
-        """ Enable standalone mode when listing fmf ids """
+        """
+        Enable standalone mode when listing fmf ids
+        """
+
         if self.opt('fmf_id'):
             return True
         return super().is_in_standalone_mode
 
-    def get_git_root(self, dir: Path) -> Path:
-        """ Find git root of the path """
+    def get_git_root(self, directory: Path) -> Path:
+        """
+        Find git root of the path
+        """
+
         output = self.run(
             Command("git", "rev-parse", "--show-toplevel"),
-            cwd=dir,
-            ignore_dry=True)
+            cwd=directory,
+            ignore_dry=True,
+        )
         assert output.stdout is not None
         return Path(output.stdout.strip("\n"))
 
-    def go(self) -> None:
-        """ Discover available tests """
-        super().go()
+    def go(self, *, logger: Optional[tmt.log.Logger] = None) -> None:
+        """
+        Discover available tests
+        """
+
+        super().go(logger=logger)
 
         # Check url and path, prepare test directory
         url = self.get('url')
@@ -309,24 +432,28 @@ class DiscoverFmf(tmt.steps.discover.DiscoverPlugin[DiscoverFmfStepData]):
         # Self checks
         if dist_git_source and not dist_git_merge and (ref or url):
             raise tmt.utils.DiscoverError(
-                "Cannot manipulate with dist-git without "
-                "the `--dist-git-merge` option.")
+                "Cannot manipulate with dist-git without the `--dist-git-merge` option."
+            )
 
         # Raise an exception if --fmf-id uses w/o url and git root
         # doesn't exist for discovered plan
         if self.opt('fmf_id'):
+
             def assert_git_url(plan_name: Optional[str] = None) -> None:
                 try:
                     subprocess.run(
                         'git rev-parse --show-toplevel'.split(),
                         stdout=subprocess.PIPE,
                         stderr=subprocess.DEVNULL,
-                        check=True)
+                        check=True,
+                    )
                 except subprocess.CalledProcessError:
                     raise tmt.utils.DiscoverError(
                         f"`tmt run discover --fmf-id` without `url` option in "
                         f"plan `{plan_name}` can be used only within"
-                        f" git repo.")
+                        f" git repo."
+                    )
+
             # It covers only one case, when there is:
             # 1) no --url on CLI
             # 2) plan w/o url exists in test run
@@ -336,7 +463,8 @@ class DiscoverFmf(tmt.steps.discover.DiscoverPlugin[DiscoverFmfStepData]):
                 except fmf.utils.RootError:
                     raise tmt.utils.DiscoverError(
                         "No metadata found in the current directory. "
-                        "Use 'tmt init' to get started.")
+                        "Use 'tmt init' to get started."
+                    )
                 for attr in fmf_tree.climb():
                     try:
                         plan_url = attr.data.get('discover').get('url')
@@ -358,12 +486,13 @@ class DiscoverFmf(tmt.steps.discover.DiscoverPlugin[DiscoverFmfStepData]):
             self.debug(f"Clone '{url}' to '{self.testdir}'.")
             # Shallow clone to speed up testing and
             # minimize data transfers if ref is not provided
-            tmt.utils.git_clone(
+            tmt.utils.git.git_clone(
                 url=url,
                 destination=self.testdir,
                 shallow=ref is None,
                 env=Environment({"GIT_ASKPASS": EnvVarValue("echo")}),
-                logger=self._logger)
+                logger=self._logger,
+            )
             git_root = self.testdir
         # Copy git repository root to workdir
         else:
@@ -372,12 +501,12 @@ class DiscoverFmf(tmt.steps.discover.DiscoverPlugin[DiscoverFmfStepData]):
             else:
                 fmf_root = Path(self.step.plan.node.root)
             requires_git = self.opt('sync-repo') or any(
-                self.get(opt) for opt in self._REQUIRES_GIT)
+                self.get(opt) for opt in self._REQUIRES_GIT
+            )
             # Path for distgit sources cannot be checked until the
             # they are extracted
             if path and not path.is_dir() and not dist_git_source:
-                raise tmt.utils.DiscoverError(
-                    f"Provided path '{path}' is not a directory.")
+                raise tmt.utils.DiscoverError(f"Provided path '{path}' is not a directory.")
             if dist_git_source:
                 # Ensure we're in a git repo when extracting dist-git sources
                 try:
@@ -385,12 +514,10 @@ class DiscoverFmf(tmt.steps.discover.DiscoverPlugin[DiscoverFmfStepData]):
                 except tmt.utils.RunError:
                     assert self.step.plan.my_run is not None  # narrow type
                     assert self.step.plan.my_run.tree is not None  # narrow type
-                    raise tmt.utils.DiscoverError(
-                        f"{self.step.plan.node.root} is not a git repo")
+                    raise tmt.utils.DiscoverError(f"{self.step.plan.node.root} is not a git repo")
             else:
                 if fmf_root is None:
-                    raise tmt.utils.DiscoverError(
-                        "No metadata found in the current directory.")
+                    raise tmt.utils.DiscoverError("No metadata found in the current directory.")
                 # Check git repository root (use fmf root if not found)
                 try:
                     git_root = self.get_git_root(fmf_root)
@@ -399,7 +526,8 @@ class DiscoverFmf(tmt.steps.discover.DiscoverPlugin[DiscoverFmfStepData]):
                     git_root = fmf_root
                 # Set path to relative path from the git root to fmf root
                 path = fmf_root.resolve().relative_to(
-                    git_root.resolve() if requires_git else fmf_root.resolve())
+                    git_root.resolve() if requires_git else fmf_root.resolve()
+                )
 
             # And finally copy the git/fmf root directory to testdir
             # (for dist-git case only when merge explicitly requested)
@@ -420,7 +548,8 @@ class DiscoverFmf(tmt.steps.discover.DiscoverPlugin[DiscoverFmfStepData]):
                 logger=self._logger,
                 workdir=self.testdir,
                 ref=ref,
-                plan=self.step.plan)
+                plan=self.step.plan,
+            )
         except tmt.utils.FileError as error:
             raise tmt.utils.DiscoverError(str(error))
 
@@ -435,9 +564,9 @@ class DiscoverFmf(tmt.steps.discover.DiscoverPlugin[DiscoverFmfStepData]):
             with contextlib.suppress(tmt.utils.RunError, AttributeError):
                 self.verbose(
                     'hash',
-                    tmt.utils.git_hash(directory=self.testdir, logger=self._logger),
-                    'green'
-                    )
+                    tmt.utils.git.git_hash(directory=self.testdir, logger=self._logger),
+                    'green',
+                )
 
         # Dist-git source processing during discover step
         if dist_git_source:
@@ -447,14 +576,15 @@ class DiscoverFmf(tmt.steps.discover.DiscoverPlugin[DiscoverFmfStepData]):
                     distgit_dir=self.testdir if ref else git_root,
                     target_dir=sourcedir,
                     handler_name=self.get('dist-git-type'),
-                    )
+                )
                 # Copy rest of files so TMT_SOURCE_DIR has patches, sources and spec file
                 # FIXME 'worktree' could be used as sourcedir when 'url' is not set
                 shutil.copytree(
                     self.testdir if ref else git_root,
                     sourcedir,
                     symlinks=True,
-                    dirs_exist_ok=True)
+                    dirs_exist_ok=True,
+                )
                 # patch & rediscover will happen later in the prepare step
                 if not self.get('dist-git-download-only'):
                     # Check if prepare is enabled, warn user if not
@@ -463,19 +593,21 @@ class DiscoverFmf(tmt.steps.discover.DiscoverPlugin[DiscoverFmfStepData]):
                     insert_to_prepare_step(
                         discover_plugin=self,
                         sourcedir=sourcedir,
-                        )
+                    )
                 # merge or not, detect later
                 self.step.plan.discover.extract_tests_later = True
                 self.info("Tests will be discovered after dist-git patching in prepare.")
                 return
             except Exception as error:
-                raise tmt.utils.DiscoverError(
-                    "Failed to process 'dist-git-source'.") from error
+                raise tmt.utils.DiscoverError("Failed to process 'dist-git-source'.") from error
         # Discover tests
         self.do_the_discovery(path)
 
     def do_the_discovery(self, path: Optional[Path] = None) -> None:
-        """ Discover the tests """
+        """
+        Discover the tests
+        """
+
         # Original path might adjusted already in go()
         if path is None:
             path = Path(cast(str, self.get('path'))) if self.get('path') else None
@@ -486,12 +618,10 @@ class DiscoverFmf(tmt.steps.discover.DiscoverPlugin[DiscoverFmfStepData]):
         else:
             self.info('path', path, 'green')
 
-        # Prepare the whole tree path and test path prefix
+        # Prepare the whole tree path
         tree_path = self.testdir / path.unrooted()
         if not tree_path.is_dir() and not self.is_dry_run:
-            raise tmt.utils.DiscoverError(
-                f"Metadata tree path '{path}' not found.")
-        prefix_path = Path('/tests') / path.unrooted()
+            raise tmt.utils.DiscoverError(f"Metadata tree path '{path}' not found.")
 
         # Show filters and test names if provided
         # Check the 'test --filter' option first, then from discover
@@ -506,45 +636,56 @@ class DiscoverFmf(tmt.steps.discover.DiscoverPlugin[DiscoverFmfStepData]):
         # Check the 'test --link' option first, then from discover
         # FIXME: cast() - typeless "dispatcher" method
         raw_link_needles = cast(list[str], tmt.Test._opt('links', []) or self.get('link', []))
-        link_needles = [tmt.base.LinkNeedle.from_spec(
-            raw_needle) for raw_needle in raw_link_needles]
+        link_needles = [
+            tmt.base.LinkNeedle.from_spec(raw_needle) for raw_needle in raw_link_needles
+        ]
 
         for link_needle in link_needles:
             self.info('link', str(link_needle), 'green')
 
-        excludes = list(
-            tmt.base.Test._opt('exclude') or self.get('exclude', []))
+        excludes = list(tmt.base.Test._opt('exclude') or self.get('exclude', []))
 
         # Filter only modified tests if requested
         modified_only = self.get('modified-only')
         modified_url = self.get('modified-url')
         if modified_url:
             previous = modified_url
-            modified_url = tmt.utils.clonable_git_url(modified_url)
+            modified_url = tmt.utils.git.clonable_git_url(modified_url)
             self.info('modified-url', modified_url, 'green')
             if previous != modified_url:
                 self.debug(f"Original url was '{previous}'.")
             self.debug(f"Fetch also '{modified_url}' as 'reference'.")
-            self.run(Command('git', 'remote', 'add', 'reference', modified_url),
-                     cwd=self.testdir)
-            self.run(Command('git', 'fetch', 'reference'), cwd=self.testdir)
+            self.run(
+                Command('git', 'remote', 'add', 'reference', modified_url),
+                cwd=self.testdir,
+            )
+            self.run(
+                Command('git', 'fetch', 'reference'),
+                cwd=self.testdir,
+            )
         if modified_only:
             modified_ref = self.get(
                 'modified-ref',
-                tmt.utils.default_branch(repository=self.testdir, logger=self._logger))
+                tmt.utils.git.default_branch(repository=self.testdir, logger=self._logger),
+            )
             self.info('modified-ref', modified_ref, 'green')
             ref_commit = self.run(
                 Command('git', 'rev-parse', '--short', str(modified_ref)),
-                cwd=self.testdir)
+                cwd=self.testdir,
+            )
             assert ref_commit.stdout is not None
             self.verbose('modified-ref hash', ref_commit.stdout.strip(), 'green')
             output = self.run(
                 Command(
                     'git', 'log', '--format=', '--stat', '--name-only', f"{modified_ref}..HEAD"
-                    ), cwd=self.testdir)
+                ),
+                cwd=self.testdir,
+            )
             if output.stdout:
-                directories = [os.path.dirname(name) for name in output.stdout.split('\n')]
-                modified = {f"^/{re.escape(name)}" for name in directories if name}
+                directories = [Path(name).parent for name in output.stdout.split('\n')]
+                modified = {
+                    f"^/{re.escape(str(directory))}" for directory in directories if directory
+                }
                 if not modified:
                     # Nothing was modified, do not select anything
                     return
@@ -562,31 +703,36 @@ class DiscoverFmf(tmt.steps.discover.DiscoverPlugin[DiscoverFmfStepData]):
         tree = tmt.Tree(
             logger=self._logger,
             path=tree_path,
-            fmf_context=self.step.plan._fmf_context)
+            fmf_context=self.step.plan._fmf_context,
+            additional_rules=self.data.adjust_tests,
+        )
         self._tests = tree.tests(
             filters=filters,
             names=names,
             conditions=["manual is False"],
             unique=False,
             links=link_needles,
-            excludes=excludes)
+            excludes=excludes,
+        )
 
         if prune:
             # Save fmf metadata
             clonedir = self.clone_dirpath / 'tests'
-            for path in tmt.utils.filter_paths(self.testdir, ['.fmf']):
+            clone_tree_path = clonedir / path.unrooted()
+            for file_path in tmt.utils.filter_paths(tree_path, [r'\.fmf']):
                 shutil.copytree(
-                    path,
-                    clonedir / path.relative_to(self.testdir),
-                    dirs_exist_ok=True)
+                    file_path,
+                    clone_tree_path / file_path.relative_to(tree_path),
+                    dirs_exist_ok=True,
+                )
 
             # Save upgrade plan
             upgrade_path = self.get('upgrade_path')
             if upgrade_path:
                 upgrade_path = f"{upgrade_path.lstrip('/')}.fmf"
-                (clonedir / upgrade_path).parent.mkdir()
-                shutil.copyfile(self.testdir / upgrade_path, clonedir / upgrade_path)
-                shutil.copymode(self.testdir / upgrade_path, clonedir / upgrade_path)
+                (clone_tree_path / upgrade_path).parent.mkdir()
+                shutil.copyfile(tree_path / upgrade_path, clone_tree_path / upgrade_path)
+                shutil.copymode(tree_path / upgrade_path, clone_tree_path / upgrade_path)
 
         # Prefix tests and handle library requires
         for test in self._tests:
@@ -598,22 +744,24 @@ class DiscoverFmf(tmt.steps.discover.DiscoverPlugin[DiscoverFmfStepData]):
                 assert test.path is not None  # narrow type
                 relative_test_path = test.path.unrooted()
                 shutil.copytree(
-                    self.testdir / relative_test_path,
-                    clonedir / relative_test_path,
-                    dirs_exist_ok=True)
+                    tree_path / relative_test_path,
+                    clone_tree_path / relative_test_path,
+                    dirs_exist_ok=True,
+                )
 
                 # Copy all parent main.fmf files
                 parent_dir = relative_test_path
                 while parent_dir.resolve() != Path.cwd().resolve():
                     parent_dir = parent_dir.parent
-                    if (self.testdir / parent_dir / 'main.fmf').exists():
+                    if (tree_path / parent_dir / 'main.fmf').exists():
                         shutil.copyfile(
-                            self.testdir / parent_dir / 'main.fmf',
-                            clonedir / parent_dir / 'main.fmf')
+                            tree_path / parent_dir / 'main.fmf',
+                            clone_tree_path / parent_dir / 'main.fmf',
+                        )
 
             # Prefix test path with 'tests' and possible 'path' prefix
             assert test.path is not None  # narrow type
-            test.path = prefix_path / test.path.unrooted()
+            test.path = Path('/tests') / path.unrooted() / test.path.unrooted()
             # Check for possible required beakerlib libraries
             if test.require or test.recommend:
                 test.require, test.recommend, _ = tmt.libraries.dependencies(
@@ -622,7 +770,8 @@ class DiscoverFmf(tmt.steps.discover.DiscoverPlugin[DiscoverFmfStepData]):
                     parent=self,
                     logger=self._logger,
                     source_location=self.testdir,
-                    target_location=clonedir if prune else self.testdir)
+                    target_location=clonedir if prune else self.testdir,
+                )
 
         if prune:
             # Clean self.testdir and copy back only required tests and files from clonedir
@@ -635,7 +784,10 @@ class DiscoverFmf(tmt.steps.discover.DiscoverPlugin[DiscoverFmfStepData]):
             shutil.rmtree(self.clone_dirpath, ignore_errors=True)
 
     def post_dist_git(self, created_content: list[Path]) -> None:
-        """ Discover tests after dist-git applied patches """
+        """
+        Discover tests after dist-git applied patches
+        """
+
         # Directory to copy out from sources
         dist_git_extract = self.get('dist-git-extract', None)
         dist_git_init = self.get('dist-git-init', False)
@@ -649,12 +801,13 @@ class DiscoverFmf(tmt.steps.discover.DiscoverPlugin[DiscoverFmfStepData]):
         # glob otherwise
         if dist_git_extract and dist_git_extract != '/':
             try:
-                dist_git_extract = Path(glob.glob(os.path.join(
-                    sourcedir, dist_git_extract.lstrip('/')))[0])
+                dist_git_extract = Path(
+                    glob.glob(str(sourcedir / dist_git_extract.lstrip('/')))[0]
+                )
             except IndexError:
                 raise tmt.utils.DiscoverError(
-                    f"Couldn't glob '{dist_git_extract}' "
-                    f"within extracted sources.")
+                    f"Couldn't glob '{dist_git_extract}' within extracted sources."
+                )
         if dist_git_init:
             if dist_git_extract == '/' or not dist_git_extract:
                 dist_git_extract = '/'
@@ -662,18 +815,18 @@ class DiscoverFmf(tmt.steps.discover.DiscoverPlugin[DiscoverFmfStepData]):
             else:
                 location = dist_git_extract
             # User specified location or 'root' of extracted sources
-            if not (Path(location) / '.fmf').is_dir() and \
-                    not self.is_dry_run:
+            if not (Path(location) / '.fmf').is_dir() and not self.is_dry_run:
                 fmf.Tree.init(location)
         elif dist_git_remove_fmf_root:
             try:
                 extracted_fmf_root = tmt.utils.find_fmf_root(
-                    sourcedir, ignore_paths=[sourcedir])[0]
+                    sourcedir,
+                    ignore_paths=[sourcedir],
+                )[0]
             except tmt.utils.MetadataError:
                 self.warn("No fmf root to remove, there isn't one already.")
             if not self.is_dry_run:
-                shutil.rmtree(os.path.join(
-                    dist_git_extract or extracted_fmf_root, '.fmf'))
+                shutil.rmtree((dist_git_extract or extracted_fmf_root) / '.fmf')
         if not dist_git_extract:
             try:
                 top_fmf_root = tmt.utils.find_fmf_root(sourcedir, ignore_paths=[sourcedir])[0]
@@ -683,7 +836,8 @@ class DiscoverFmf(tmt.steps.discover.DiscoverPlugin[DiscoverFmfStepData]):
                     self.warn(
                         "Extracted sources do not contain fmf root, "
                         "merging with plan data. Avoid this warning by "
-                        "explicit use of the '--dist-git-merge' option.")
+                        "explicit use of the '--dist-git-merge' option."
+                    )
                     # FIXME - Deprecate this behavior?
                     git_root = self.get_git_root(Path(self.step.plan.node.root))
                     self.debug(f"Copy '{git_root}' to '{self.testdir}'.")
@@ -707,7 +861,8 @@ class DiscoverFmf(tmt.steps.discover.DiscoverPlugin[DiscoverFmfStepData]):
                         sourcedir / to_copy,
                         self.testdir if flatten else self.testdir / to_copy,
                         symlinks=True,
-                        dirs_exist_ok=True)
+                        dirs_exist_ok=True,
+                    )
                 else:
                     shutil.copyfile(src, self.testdir / to_copy)
 
@@ -723,7 +878,9 @@ class DiscoverFmf(tmt.steps.discover.DiscoverPlugin[DiscoverFmfStepData]):
         # Prefix test name only if multiple plugins configured
         prefix = f'/{self.name}' if len(self.step.phases()) > 1 else ''
         # Check discovered tests, modify test name/path
-        for test in self.tests(enabled=True):
+        for test_origin in self.tests(enabled=True):
+            test = test_origin.test
+
             test.name = f"{prefix}{test.name}"
             test.path = Path(f"/{self.safe_name}{test.path}")
             # Update test environment with plan environment
@@ -734,16 +891,22 @@ class DiscoverFmf(tmt.steps.discover.DiscoverPlugin[DiscoverFmfStepData]):
         self.step.summary()
 
     def tests(
-            self,
-            *,
-            phase_name: Optional[str] = None,
-            enabled: Optional[bool] = None) -> list['tmt.Test']:
-        """ Return all discovered tests """
+        self, *, phase_name: Optional[str] = None, enabled: Optional[bool] = None
+    ) -> list[tmt.steps.discover.TestOrigin]:
+        """
+        Return all discovered tests
+        """
 
         if phase_name is not None and phase_name != self.name:
             return []
 
         if enabled is None:
-            return self._tests
+            return [
+                tmt.steps.discover.TestOrigin(test=test, phase=self.name) for test in self._tests
+            ]
 
-        return [test for test in self._tests if test.enabled is enabled]
+        return [
+            tmt.steps.discover.TestOrigin(test=test, phase=self.name)
+            for test in self._tests
+            if test.enabled is enabled
+        ]
