@@ -1,4 +1,3 @@
-import dataclasses
 import re
 import shutil
 from collections.abc import Iterator
@@ -15,6 +14,7 @@ import tmt.package_managers
 import tmt.steps
 import tmt.steps.prepare
 import tmt.utils
+from tmt.container import container, field
 from tmt.package_managers import (
     FileSystemPath,
     Installable,
@@ -22,9 +22,11 @@ from tmt.package_managers import (
     Package,
     PackagePath,
     PackageUrl,
-    )
+)
+from tmt.package_managers.bootc import Bootc, BootcEngine
+from tmt.result import PhaseResult
 from tmt.steps.provision import Guest
-from tmt.utils import Command, Path, ShellScript, field
+from tmt.utils import Command, Path, ShellScript
 
 COPR_URL = 'https://copr.fedorainfracloud.org/coprs'
 
@@ -33,7 +35,9 @@ T = TypeVar('T')
 
 
 class InstallBase(tmt.utils.Common):
-    """ Base class for installation implementations """
+    """
+    Base class for installation implementations
+    """
 
     guest: Guest
 
@@ -48,16 +52,20 @@ class InstallBase(tmt.utils.Common):
     package_directory: Path
 
     def __init__(
-            self,
-            *,
-            parent: 'PrepareInstall',
-            guest: Guest,
-            dependencies: list[tmt.base.DependencySimple],
-            directories: list[Path],
-            exclude: list[str],
-            logger: tmt.log.Logger,
-            **kwargs: Any) -> None:
-        """ Initialize installation data """
+        self,
+        *,
+        parent: 'PrepareInstall',
+        guest: Guest,
+        dependencies: list[tmt.base.DependencySimple],
+        directories: list[Path],
+        exclude: list[str],
+        logger: tmt.log.Logger,
+        **kwargs: Any,
+    ) -> None:
+        """
+        Initialize installation data
+        """
+
         super().__init__(logger=logger, parent=parent, relative_indent=0, guest=guest, **kwargs)
 
         if not dependencies and not directories:
@@ -72,10 +80,14 @@ class InstallBase(tmt.utils.Common):
         self.prepare_installables(dependencies, directories)
 
     def prepare_installables(
-            self,
-            dependencies: list[tmt.base.DependencySimple],
-            directories: list[Path]) -> None:
-        """ Process package names and directories """
+        self,
+        dependencies: list[tmt.base.DependencySimple],
+        directories: list[Path],
+    ) -> None:
+        """
+        Process package names and directories
+        """
+
         self.packages = []
         self.local_packages = []
         self.remote_packages = []
@@ -91,11 +103,8 @@ class InstallBase(tmt.utils.Common):
                 # Strip the '-debuginfo' string from package name
                 # (installing with it doesn't work on RHEL7)
                 self.debuginfo_packages.append(
-                    Package(
-                        re.sub(
-                            r"-debuginfo((?=\.)|$)",
-                            "",
-                            str(dependency))))
+                    Package(re.sub(r"-debuginfo((?=\.)|$)", "", str(dependency)))
+                )
 
             elif dependency.startswith('/'):
                 self.packages.append(FileSystemPath(dependency))
@@ -114,11 +123,16 @@ class InstallBase(tmt.utils.Common):
                     self.local_packages.append(PackagePath(filepath))
 
     def prepare_repository(self) -> None:
-        """ Configure additional repository """
+        """
+        Configure additional repository
+        """
+
         raise NotImplementedError
 
     def list_installables(self, title: str, *installables: Installable) -> Iterator[Installable]:
-        """ Show package info and return package names """
+        """
+        Show package info and return package names
+        """
 
         # Show a brief summary by default
         if not self.verbosity_level:
@@ -134,7 +148,10 @@ class InstallBase(tmt.utils.Common):
         yield from installables
 
     def prepare_install_local(self) -> None:
-        """ Copy packages to the test system """
+        """
+        Copy packages to the test system
+        """
+
         assert self.parent is not None
         # FIXME: cast() - https://github.com/teemtee/tmt/issues/1372
         workdir = cast(PrepareInstall, self.parent).step.workdir
@@ -151,23 +168,51 @@ class InstallBase(tmt.utils.Common):
         self.guest.push()
 
     def install_from_repository(self) -> None:
-        """ Default base install method for packages from repositories """
-        pass
+        """
+        Default base install method for packages from repositories
+        """
 
     def install_local(self) -> None:
-        """ Default base install method for local packages """
-        pass
+        """
+        Default base install method for local packages
+        """
 
     def install_from_url(self) -> None:
-        """ Default base install method for packages which are from URL """
-        pass
+        """
+        Default base install method for packages which are from URL
+        """
 
     def install_debuginfo(self) -> None:
-        """ Default base install method for debuginfo packages """
-        pass
+        """
+        Default base install method for debuginfo packages
+        """
 
     def install(self) -> None:
-        """ Perform the actual installation """
+        """
+        Perform the actual installation
+        """
+
+        try:
+            self._install()
+
+        except Exception as exc1:
+            # We do not have any special handling for exceptions raised by the following code.
+            # Wrapping them with try/except gives us a chance to attach the original exception
+            # to whatever the code may raise, and therefore preserve the information attached
+            # to the original exception.
+            try:
+                # Refresh cache in case of recent but not updated change do repodata
+                self.guest.package_manager.refresh_metadata()
+                self._install()
+
+            except Exception as exc2:
+                raise exc2 from exc1
+
+    def _install(self) -> None:
+        """
+        Helper method to perform the actual installation steps
+        """
+
         if self.local_packages:
             self.prepare_install_local()
             self.install_local()
@@ -181,8 +226,10 @@ class InstallBase(tmt.utils.Common):
     def rpm_check(self, package: str, mode: str = '-q') -> None:
         """
         Run rpm command to check package existence
+
         Throws tmt.utils.RunError
         """
+
         output = self.guest.execute(Command('rpm', mode, package), silent=True)
         assert output.stdout
         self.debug(f"Package '{output.stdout.strip()}' already installed.")
@@ -200,7 +247,10 @@ class Copr(tmt.utils.Common):
         self.guest = guest
 
     def enable_copr_epel6(self, copr: str) -> None:
-        """ Manually enable copr repositories for epel6 """
+        """
+        Manually enable copr repositories for epel6
+        """
+
         # Parse the copr repo name
         matched = re.match("^(@)?([^/]+)/([^/]+)$", copr)
         if not matched:
@@ -214,24 +264,30 @@ class Copr(tmt.utils.Common):
         url = '/'.join(parts)
         # Download the repo file on guest
         try:
-            self.guest.execute(Command('curl', '-LOf', url),
-                               cwd=Path('/etc/yum.repos.d'), silent=True)
+            self.guest.execute(
+                Command('curl', '-LOf', url),
+                cwd=Path('/etc/yum.repos.d'),
+                silent=True,
+            )
         except tmt.utils.RunError as error:
             if error.stderr and 'not found' in error.stderr.lower():
-                raise tmt.utils.PrepareError(
-                    f"Copr repository '{copr}' not found.")
+                raise tmt.utils.PrepareError(f"Copr repository '{copr}' not found.")
             raise
 
     def enable_copr(self, repositories: list[str]) -> None:
-        """ Enable requested copr repositories """
+        """
+        Enable requested copr repositories
+        """
 
         if not repositories:
             return
 
+        package_manager = self.guest.package_manager
+
         # Try to install copr plugin
         self.debug('Make sure the copr plugin is available.')
         try:
-            self.guest.package_manager.install(Package(self.copr_plugin))
+            package_manager.install(Package(self.copr_plugin))
 
         # Enable repositories manually for epel6
         except tmt.utils.RunError:
@@ -244,101 +300,122 @@ class Copr(tmt.utils.Common):
             for repository in repositories:
                 self.info('copr', repository, 'green')
 
-                self.guest.execute(ShellScript(
-                    f"{self.guest.package_manager.command.to_script()} copr "
-                    f"{self.guest.package_manager.options.to_script()} enable -y {repository}"
-                    ))
+                self.guest.execute(
+                    ShellScript(
+                        f"{package_manager.engine.command.to_script()} copr "
+                        f"{package_manager.engine.options.to_script()} enable -y {repository}"
+                    )
+                )
 
 
 class InstallDnf(InstallBase, Copr):
-    """ Install packages using dnf """
+    """
+    Install packages using dnf
+    """
 
     copr_plugin = "dnf-plugins-core"
 
     def install_local(self) -> None:
-        """ Install packages stored in a local directory """
+        """
+        Install packages stored in a local directory
+        """
 
         # Use both dnf install/reinstall to get all packages refreshed
         # FIXME Simplify this once BZ#1831022 is fixed/implemented.
-        filelist = [PackagePath(self.package_directory / filename)
-                    for filename in self.local_packages]
+        filelist = [
+            PackagePath(self.package_directory / filename.name) for filename in self.local_packages
+        ]
 
         self.guest.package_manager.install(
             *filelist,
             options=Options(
                 excluded_packages=self.exclude,
                 skip_missing=self.skip_missing,
-                check_first=False
-                )
-            )
+                check_first=False,
+            ),
+        )
 
         self.guest.package_manager.reinstall(
             *filelist,
             options=Options(
                 excluded_packages=self.exclude,
                 skip_missing=self.skip_missing,
-                check_first=False
-                )
-            )
+                check_first=False,
+            ),
+        )
 
         summary = fmf.utils.listed([str(path) for path in self.local_packages], 'local package')
         self.info('total', f"{summary} installed", 'green')
 
     def install_from_url(self) -> None:
-        """ Install packages stored on a remote URL """
+        """
+        Install packages stored on a remote URL
+        """
 
         self.guest.package_manager.install(
             *self.list_installables("remote package", *self.remote_packages),
             options=Options(
                 excluded_packages=self.exclude,
-                skip_missing=self.skip_missing
-                )
-            )
+                skip_missing=self.skip_missing,
+            ),
+        )
 
     def install_from_repository(self) -> None:
-        """ Install packages from a repository """
+        """
+        Install packages from a repository
+        """
 
         self.guest.package_manager.install(
             *self.list_installables("package", *self.packages),
             options=Options(
                 excluded_packages=self.exclude,
-                skip_missing=self.skip_missing
-                )
-            )
+                skip_missing=self.skip_missing,
+            ),
+        )
 
     def install_debuginfo(self) -> None:
-        """ Install debuginfo packages """
+        """
+        Install debuginfo packages
+        """
+
         packages = self.list_installables("debuginfo", *self.debuginfo_packages)
 
         self.guest.package_manager.install_debuginfo(
             *packages,
             options=Options(
                 excluded_packages=self.exclude,
-                skip_missing=self.skip_missing
-                )
-            )
+                skip_missing=self.skip_missing,
+            ),
+        )
 
         # Check the packages are installed on the guest because 'debuginfo-install'
         # returns 0 even though it didn't manage to install the required packages
         if not self.skip_missing:
             self.guest.package_manager.check_presence(
-                *[Package(f'{package}-debuginfo') for package in self.debuginfo_packages])
+                *[Package(f'{package}-debuginfo') for package in self.debuginfo_packages]
+            )
 
 
 class InstallDnf5(InstallDnf):
-    """ Install packages using dnf5 """
+    """
+    Install packages using dnf5
+    """
 
     copr_plugin = "dnf5-command(copr)"
 
 
 class InstallYum(InstallDnf):
-    """ Install packages using yum """
+    """
+    Install packages using yum
+    """
 
     copr_plugin = "yum-plugin-copr"
 
 
 class InstallRpmOstree(InstallBase, Copr):
-    """ Install packages using rpm-ostree """
+    """
+    Install packages using rpm-ostree
+    """
 
     copr_plugin = "dnf-plugins-core"
 
@@ -346,7 +423,9 @@ class InstallRpmOstree(InstallBase, Copr):
     required_packages: list[Union[Package, FileSystemPath]]
 
     def enable_copr(self, repositories: list[str]) -> None:
-        """ Enable requested copr repositories """
+        """
+        Enable requested copr repositories
+        """
 
         # rpm-ostree can step outside of its zone of competence, and use
         # another package manager to enable copr. We just need to cheat
@@ -361,11 +440,15 @@ class InstallRpmOstree(InstallBase, Copr):
             dependencies=[],
             directories=[],
             exclude=[],
-            logger=self._logger).enable_copr(repositories)
+            logger=self._logger,
+        ).enable_copr(repositories)
         self.guest.facts.package_manager = 'rpm-ostree'
 
     def sort_packages(self) -> None:
-        """ Identify required and recommended packages """
+        """
+        Identify required and recommended packages
+        """
+
         self.recommended_packages = []
         self.required_packages = []
         for package in self.packages:
@@ -378,20 +461,24 @@ class InstallRpmOstree(InstallBase, Copr):
                     self.required_packages.append(package)
 
     def install_debuginfo(self) -> None:
-        """ Install debuginfo packages """
+        """
+        Install debuginfo packages
+        """
+
         self.warn("Installation of debuginfo packages not supported yet.")
 
     def install_local(self) -> None:
-        """ Install copied local packages """
+        """
+        Install copied local packages
+        """
+
         local_packages_installed: list[PackagePath] = []
         for package in self.local_packages:
             try:
                 self.guest.package_manager.install(
                     PackagePath(self.package_directory / package.name),
-                    options=Options(
-                        check_first=False
-                        )
-                    )
+                    options=Options(check_first=False),
+                )
                 local_packages_installed.append(package)
             except tmt.utils.RunError as error:
                 self.warn(f"Local package '{package}' not installed: {error.stderr}")
@@ -399,7 +486,10 @@ class InstallRpmOstree(InstallBase, Copr):
         self.info('total', f"{summary} installed", 'green')
 
     def install_from_repository(self) -> None:
-        """ Install packages from the repository """
+        """
+        Install packages from the repository
+        """
+
         self.sort_packages()
 
         # Install recommended packages
@@ -416,79 +506,174 @@ class InstallRpmOstree(InstallBase, Copr):
         # Install required packages
         if self.required_packages:
             self.guest.package_manager.install(
-                *self.list_installables("package", *self.required_packages))
+                *self.list_installables("package", *self.required_packages)
+            )
 
 
 class InstallApt(InstallBase):
-    """ Install packages using apt """
+    """
+    Install packages using apt
+    """
 
     def install_local(self) -> None:
-        """ Install packages stored in a local directory """
+        """
+        Install packages stored in a local directory
+        """
 
-        filelist = [PackagePath(self.package_directory / filename)
-                    for filename in self.local_packages]
+        filelist = [
+            PackagePath(self.package_directory / filename) for filename in self.local_packages
+        ]
 
         self.guest.package_manager.install(
             *self.list_installables('local packages', *filelist),
             options=Options(
-                excluded_packages=self.exclude,
-                skip_missing=self.skip_missing,
-                check_first=False
-                )
-            )
+                excluded_packages=self.exclude, skip_missing=self.skip_missing, check_first=False
+            ),
+        )
 
         summary = fmf.utils.listed([str(path) for path in self.local_packages], 'local package')
         self.info('total', f"{summary} installed", 'green')
 
     def install_from_url(self) -> None:
-        """ Install packages stored on a remote URL """
+        """
+        Install packages stored on a remote URL
+        """
 
         self.guest.package_manager.install(
             *self.list_installables("remote package", *self.remote_packages),
             options=Options(
                 excluded_packages=self.exclude,
-                skip_missing=self.skip_missing
-                )
-            )
+                skip_missing=self.skip_missing,
+            ),
+        )
 
     def install_from_repository(self) -> None:
-        """ Install packages from a repository """
+        """
+        Install packages from a repository
+        """
 
         self.guest.package_manager.install(
             *self.list_installables("package", *self.packages),
             options=Options(
                 excluded_packages=self.exclude,
-                skip_missing=self.skip_missing
-                )
-            )
+                skip_missing=self.skip_missing,
+            ),
+        )
 
     def install_debuginfo(self) -> None:
-        """ Install debuginfo packages """
+        """
+        Install debuginfo packages
+        """
+
         packages = self.list_installables("debuginfo", *self.debuginfo_packages)
 
         self.guest.package_manager.install_debuginfo(
             *packages,
             options=Options(
                 excluded_packages=self.exclude,
-                skip_missing=self.skip_missing
-                )
-            )
+                skip_missing=self.skip_missing,
+            ),
+        )
 
         # Check the packages are installed on the guest because 'debuginfo-install'
         # returns 0 even though it didn't manage to install the required packages
         if not self.skip_missing:
             self.guest.package_manager.check_presence(
-                *[Package(f'{package}-debuginfo') for package in self.debuginfo_packages])
+                *[Package(f'{package}-debuginfo') for package in self.debuginfo_packages]
+            )
+
+
+class InstallBootc(InstallBase):
+    """Install packages using bootc container image mode"""
+
+    def install_from_repository(self) -> None:
+        self.guest.package_manager.engine.install(
+            *self.list_installables("package", *self.packages),
+            options=Options(
+                excluded_packages=self.exclude,
+                skip_missing=self.skip_missing,
+            ),
+        )
+
+    def install_from_url(self) -> None:
+        self.guest.package_manager.engine.install(
+            *self.list_installables("remote package", *self.remote_packages),
+            options=Options(
+                excluded_packages=self.exclude,
+                skip_missing=self.skip_missing,
+            ),
+        )
+
+    def install_local(self) -> None:
+        # Filelist for packages on the guest
+        filelist = [
+            PackagePath(self.package_directory / filename.name) for filename in self.local_packages
+        ]
+
+        cast(BootcEngine, self.guest.package_manager.engine).containerfile_directives.append(
+            f'RUN mkdir -p {self.package_directory}'
+        )
+        cast(BootcEngine, self.guest.package_manager.engine).containerfile_directives.append(
+            f'COPY {" ".join(str(file) for file in filelist)} {self.package_directory}'
+        )
+
+        self.guest.package_manager.engine.install(
+            *filelist,
+            options=Options(
+                excluded_packages=self.exclude,
+                skip_missing=self.skip_missing,
+                check_first=False,
+            ),
+        )
+
+        self.guest.package_manager.engine.reinstall(
+            *filelist,
+            options=Options(
+                excluded_packages=self.exclude,
+                skip_missing=self.skip_missing,
+                check_first=False,
+            ),
+        )
+
+    def install_debuginfo(self) -> None:
+        packages = self.list_installables("debuginfo", *self.debuginfo_packages)
+
+        self.guest.package_manager.engine.install_debuginfo(
+            *packages,
+            options=Options(
+                excluded_packages=self.exclude,
+                skip_missing=self.skip_missing,
+            ),
+        )
+
+        # Check the packages are installed on the guest because 'debuginfo-install'
+        # returns 0 even though it didn't manage to install the required packages
+        if not self.skip_missing:
+            self.guest.package_manager.engine.check_presence(
+                *[Package(f'{package}-debuginfo') for package in self.debuginfo_packages]
+            )
+
+    def _install(self) -> None:
+        """Coordinate installation process through containerfile building and switching"""
+        # Call base install methods to collect all package types
+        super()._install()
+
+        cast(Bootc, self.guest.package_manager).build_container()
 
 
 class InstallApk(InstallBase):
-    """ Install packages using apk """
+    """
+    Install packages using apk
+    """
 
     def install_local(self) -> None:
-        """ Install packages stored in a local directory """
+        """
+        Install packages stored in a local directory
+        """
 
-        filelist = [PackagePath(self.package_directory / filename)
-                    for filename in self.local_packages]
+        filelist = [
+            PackagePath(self.package_directory / filename) for filename in self.local_packages
+        ]
 
         self.guest.package_manager.install(
             *self.list_installables('local packages', *filelist),
@@ -496,40 +681,48 @@ class InstallApk(InstallBase):
                 excluded_packages=self.exclude,
                 skip_missing=self.skip_missing,
                 allow_untrusted=True,
-                check_first=False
-                )
-            )
+                check_first=False,
+            ),
+        )
 
         summary = fmf.utils.listed([str(path) for path in self.local_packages], 'local package')
         self.info('total', f"{summary} installed", 'green')
 
     def install_from_url(self) -> None:
-        """ Install packages stored on a remote URL """
+        """
+        Install packages stored on a remote URL
+        """
 
         raise tmt.utils.PrepareError(
             f'Package manager "{self.guest.facts.package_manager}" '
-            'does not support installing from a remote URL.')
+            'does not support installing from a remote URL.'
+        )
 
     def install_from_repository(self) -> None:
-        """ Install packages from a repository """
+        """
+        Install packages from a repository
+        """
 
         self.guest.package_manager.install(
             *self.list_installables("package", *self.packages),
             options=Options(
                 excluded_packages=self.exclude,
-                skip_missing=self.skip_missing
-                )
-            )
+                skip_missing=self.skip_missing,
+            ),
+        )
 
     def install_debuginfo(self) -> None:
-        """ Install debuginfo packages """
+        """
+        Install debuginfo packages
+        """
 
         raise tmt.utils.PrepareError(
             f'Package manager "{self.guest.facts.package_manager}" does not support '
-            'installing debuginfo packages.')
+            'installing debuginfo packages.'
+        )
 
 
-@dataclasses.dataclass
+@container
 class PrepareInstallData(tmt.steps.prepare.PrepareStepData):
     package: list[tmt.base.DependencySimple] = field(
         default_factory=list,
@@ -541,13 +734,13 @@ class PrepareInstallData(tmt.steps.prepare.PrepareStepData):
         normalize=lambda key_address, value, logger: tmt.base.assert_simple_dependencies(
             tmt.base.normalize_require(key_address, value, logger),
             "'install' plugin support simple packages only, no fmf links are allowed",
-            logger),
+            logger,
+        ),
         serialize=lambda packages: [package.to_spec() for package in packages],
         unserialize=lambda serialized: [
-            tmt.base.DependencySimple.from_spec(package)
-            for package in serialized
-            ]
-        )
+            tmt.base.DependencySimple.from_spec(package) for package in serialized
+        ],
+    )
 
     directory: list[Path] = field(
         default_factory=list,
@@ -555,8 +748,8 @@ class PrepareInstallData(tmt.steps.prepare.PrepareStepData):
         metavar='PATH',
         multiple=True,
         help='Path to a local directory with rpm packages.',
-        normalize=tmt.utils.normalize_path_list
-        )
+        normalize=tmt.utils.normalize_path_list,
+    )
 
     copr: list[str] = field(
         default_factory=list,
@@ -564,8 +757,8 @@ class PrepareInstallData(tmt.steps.prepare.PrepareStepData):
         metavar='REPO',
         multiple=True,
         help='Copr repository to be enabled.',
-        normalize=tmt.utils.normalize_string_list
-        )
+        normalize=tmt.utils.normalize_string_list,
+    )
 
     exclude: list[str] = field(
         default_factory=list,
@@ -573,8 +766,8 @@ class PrepareInstallData(tmt.steps.prepare.PrepareStepData):
         metavar='PACKAGE',
         multiple=True,
         help='Packages to be skipped during installation.',
-        normalize=tmt.utils.normalize_string_list
-        )
+        normalize=tmt.utils.normalize_string_list,
+    )
 
     # TODO: use enum
     missing: Literal['skip', 'fail'] = field(
@@ -582,8 +775,8 @@ class PrepareInstallData(tmt.steps.prepare.PrepareStepData):
         option=('-m', '--missing'),
         metavar='ACTION',
         choices=['fail', 'skip'],
-        help='Action on missing packages, fail (default) or skip.'
-        )
+        help='Action on missing packages, fail (default) or skip.',
+    )
 
 
 @tmt.steps.provides_method('install')
@@ -601,23 +794,57 @@ class PrepareInstall(tmt.steps.prepare.PreparePlugin[PrepareInstallData]):
             package: tmt-all
             missing: fail
 
-    Use ``copr`` for enabling desired Copr repository and ``missing`` to choose
+    Use ``copr`` for enabling a desired Copr repository and ``missing`` to choose
     whether missing packages should be silently ignored (``skip``) or a
     preparation error should be reported (``fail``), which is the default.
 
-    In addition to package name you can also use one or more paths to
-    local rpm files to be installed:
+    One or more RPM packages can be specified under the
+    ``package`` attribute. The packages will be installed
+    on the guest. They can either be specified using their
+    names, paths to local rpm files or urls to remote rpms.
 
     .. code-block:: yaml
 
+        # Install local rpms using file path
         prepare:
             how: install
             package:
                 - tmp/RPMS/noarch/tmt-0.15-1.fc31.noarch.rpm
                 - tmp/RPMS/noarch/python3-tmt-0.15-1.fc31.noarch.rpm
 
+    .. code-block:: yaml
+
+        # Install remote packages using url
+        prepare:
+            how: install
+            package:
+              - https://dl.fedoraproject.org/pub/epel/epel-release-latest-8.noarch.rpm
+              - https://dl.fedoraproject.org/pub/epel/epel-next-release-latest-8.noarch.rpm
+
+    .. code-block:: yaml
+
+        # Install the whole directory, exclude selected packages
+        prepare:
+            how: install
+            directory:
+              - tmp/RPMS/noarch
+            exclude:
+              - tmt+all
+              - tmt+provision-virtual
+
+    .. code-block:: yaml
+
+        prepare:
+            how: install
+            # Repository with a group owner (@ prefixed) requires quotes, e.g.
+            # copr: "@osci/rpminspect"
+            copr: psss/tmt
+            package: tmt-all
+            missing: skip
+
     Use ``directory`` to install all packages from given folder and
-    ``exclude`` to skip selected packages:
+    ``exclude`` to skip selected packages (globbing characters are supported as
+    well).
 
     .. code-block:: yaml
 
@@ -640,17 +867,21 @@ class PrepareInstall(tmt.steps.prepare.PreparePlugin[PrepareInstallData]):
     _data_class = PrepareInstallData
 
     def go(
-            self,
-            *,
-            guest: 'Guest',
-            environment: Optional[tmt.utils.Environment] = None,
-            logger: tmt.log.Logger) -> None:
-        """ Perform preparation for the guests """
-        super().go(guest=guest, environment=environment, logger=logger)
+        self,
+        *,
+        guest: 'Guest',
+        environment: Optional[tmt.utils.Environment] = None,
+        logger: tmt.log.Logger,
+    ) -> list[PhaseResult]:
+        """
+        Perform preparation for the guests
+        """
+
+        results = super().go(guest=guest, environment=environment, logger=logger)
 
         # Nothing to do in dry mode
         if self.is_dry_run:
-            return
+            return results
 
         # Pick the right implementation
         # TODO: it'd be nice to use a "plugin registry" and make the
@@ -658,14 +889,25 @@ class PrepareInstall(tmt.steps.prepare.PreparePlugin[PrepareInstallData]):
         # shipped as plugins, but we still need a matching *installation* class.
         # But do we really need a class per package manager family? Maybe the
         # code could be integrated into package manager plugins directly.
-        if guest.facts.package_manager == 'rpm-ostree':
-            installer: InstallBase = InstallRpmOstree(
+        if guest.facts.package_manager == 'bootc':
+            installer: InstallBase = InstallBootc(
                 logger=logger,
                 parent=self,
                 dependencies=self.data.package,
                 directories=self.data.directory,
                 exclude=self.data.exclude,
-                guest=guest)
+                guest=guest,
+            )
+
+        elif guest.facts.package_manager == 'rpm-ostree':
+            installer = InstallRpmOstree(
+                logger=logger,
+                parent=self,
+                dependencies=self.data.package,
+                directories=self.data.directory,
+                exclude=self.data.exclude,
+                guest=guest,
+            )
 
         elif guest.facts.package_manager == 'dnf5':
             installer = InstallDnf5(
@@ -674,7 +916,8 @@ class PrepareInstall(tmt.steps.prepare.PreparePlugin[PrepareInstallData]):
                 dependencies=self.data.package,
                 directories=self.data.directory,
                 exclude=self.data.exclude,
-                guest=guest)
+                guest=guest,
+            )
 
         elif guest.facts.package_manager == 'dnf':
             installer = InstallDnf(
@@ -683,7 +926,8 @@ class PrepareInstall(tmt.steps.prepare.PreparePlugin[PrepareInstallData]):
                 dependencies=self.data.package,
                 directories=self.data.directory,
                 exclude=self.data.exclude,
-                guest=guest)
+                guest=guest,
+            )
 
         elif guest.facts.package_manager == 'yum':
             installer = InstallYum(
@@ -692,7 +936,8 @@ class PrepareInstall(tmt.steps.prepare.PreparePlugin[PrepareInstallData]):
                 dependencies=self.data.package,
                 directories=self.data.directory,
                 exclude=self.data.exclude,
-                guest=guest)
+                guest=guest,
+            )
 
         elif guest.facts.package_manager == 'apt':
             installer = InstallApt(
@@ -701,7 +946,8 @@ class PrepareInstall(tmt.steps.prepare.PreparePlugin[PrepareInstallData]):
                 dependencies=self.data.package,
                 directories=self.data.directory,
                 exclude=self.data.exclude,
-                guest=guest)
+                guest=guest,
+            )
 
         elif guest.facts.package_manager == 'apk':
             installer = InstallApk(
@@ -710,7 +956,8 @@ class PrepareInstall(tmt.steps.prepare.PreparePlugin[PrepareInstallData]):
                 dependencies=self.data.package,
                 directories=self.data.directory,
                 exclude=self.data.exclude,
-                guest=guest)
+                guest=guest,
+            )
 
         elif guest.facts.package_manager is None:
             raise tmt.utils.PrepareError('Unrecognized package manager.')
@@ -718,7 +965,8 @@ class PrepareInstall(tmt.steps.prepare.PreparePlugin[PrepareInstallData]):
         else:
             raise tmt.utils.PrepareError(
                 f"Package manager '{guest.facts.package_manager}' "
-                "is not supported by 'prepare/install'.")
+                "is not supported by 'prepare/install'."
+            )
 
         # Enable copr repositories...
         if isinstance(installer, Copr):
@@ -726,3 +974,5 @@ class PrepareInstall(tmt.steps.prepare.PreparePlugin[PrepareInstallData]):
 
         # ... and install packages.
         installer.install()
+
+        return results
