@@ -1,6 +1,6 @@
-import dataclasses
 import datetime
-from typing import Any, Optional, TypedDict, cast
+import functools
+from typing import Any, Optional, TypedDict, Union, cast
 
 import requests
 
@@ -11,20 +11,30 @@ import tmt.options
 import tmt.steps
 import tmt.steps.provision
 import tmt.utils
+import tmt.utils.signals
+from tmt.container import container, field
 from tmt.utils import (
+    Command,
     ProvisionError,
+    ShellScript,
     UpdatableMessage,
-    cached_property,
     dict_to_yaml,
-    field,
+    normalize_string_dict,
     retry_session,
-    )
+)
 
 # List of Artemis API versions supported and understood by this plugin.
 # Since API gains support for new features over time, it is important to
 # know when particular feature became available, and avoid using it with
 # older APIs.
 SUPPORTED_API_VERSIONS = (
+    # NEW: guest reboot
+    '0.0.74',
+    # NEW: allow passing security group rules for guest creation
+    '0.0.72',
+    # NEW: guest log API adds multiple blobs
+    # NEW: dropped boot.method enum
+    '0.0.70',
     # NEW: zcrypt HW requirement
     '0.0.69',
     # NEW: added cpu.flag constraint
@@ -50,8 +60,8 @@ SUPPORTED_API_VERSIONS = (
     # NEW: boot.method HW constraint
     '0.0.32',
     # NEW: network HW constraint
-    '0.0.28'
-    )
+    '0.0.28',
+)
 
 
 # TODO: Artemis does not have any whoami endpoint which would report
@@ -60,8 +70,8 @@ SUPPORTED_LOG_TYPES = [
     'console:dump/blob',
     'console:dump/url',
     'console:interactive/url',
-    'sys.log:dump/url'
-    ]
+    'sys.log:dump/url',
+]
 
 
 # The default Artemis API version - the most recent supported versions
@@ -81,109 +91,87 @@ DEFAULT_API_RETRIES = 10
 DEFAULT_RETRY_BACKOFF_FACTOR = 1
 
 
-def _normalize_user_data(
-        key_address: str,
-        raw_value: Any,
-        logger: tmt.log.Logger) -> dict[str, str]:
-    if isinstance(raw_value, dict):
-        return {
-            str(key).strip(): str(value).strip() for key, value in raw_value.items()
-            }
-
-    if isinstance(raw_value, (list, tuple)):
-        user_data = {}
-
-        for datum in raw_value:
-            try:
-                key, value = datum.split('=', 1)
-
-            except ValueError as exc:
-                raise tmt.utils.NormalizationError(
-                    key_address, datum, 'a KEY=VALUE string') from exc
-
-            user_data[key.strip()] = value.strip()
-
-        return user_data
-
-    raise tmt.utils.NormalizationError(
-        key_address, value, 'a dictionary or a list of KEY=VALUE strings')
-
-
-def _normalize_log_type(
-        key_address: str,
-        raw_value: Any,
-        logger: tmt.log.Logger) -> list[str]:
+def _normalize_log_type(key_address: str, raw_value: Any, logger: tmt.log.Logger) -> list[str]:
     if isinstance(raw_value, str):
         return [raw_value]
 
     if isinstance(raw_value, (list, tuple)):
         return [str(item) for item in raw_value]
 
-    raise tmt.utils.NormalizationError(
-        key_address, raw_value, 'a string or a list of strings')
+    raise tmt.utils.NormalizationError(key_address, raw_value, 'a string or a list of strings')
 
 
-@dataclasses.dataclass
+@container
 class ArtemisGuestData(tmt.steps.provision.GuestSshData):
     user: str = field(
         default=DEFAULT_USER,
         option=('-u', '--user'),
         metavar='USERNAME',
-        help='Username to use for all guest operations.')
+        help='Username to use for all guest operations.',
+    )
 
     # API
     api_url: str = field(
         default=DEFAULT_API_URL,
         option='--api-url',
         metavar='URL',
-        help="Artemis API URL.")
+        help="Artemis API URL.",
+    )
     api_version: str = field(
         default=DEFAULT_API_VERSION,
         option='--api-version',
         metavar='X.Y.Z',
         help="Artemis API version to use.",
-        choices=SUPPORTED_API_VERSIONS)
+        choices=SUPPORTED_API_VERSIONS,
+    )
 
     # Guest request properties
     arch: str = field(
         default=DEFAULT_ARCH,
         option='--arch',
         metavar='ARCH',
-        help='Architecture to provision.')
+        help='Architecture to provision.',
+    )
     image: Optional[str] = field(
         default=None,
         option=('-i', '--image'),
         metavar='COMPOSE',
-        help='Image (or "compose" in Artemis terminology) to provision.')
+        help='Image (or "compose" in Artemis terminology) to provision.',
+    )
     pool: Optional[str] = field(
         default=None,
         option='--pool',
         metavar='NAME',
-        help='Pool to enforce.')
+        help='Pool to enforce.',
+    )
     priority_group: str = field(
         default=DEFAULT_PRIORITY_GROUP,
         option='--priority-group',
         metavar='NAME',
-        help='Provisioning priority group.')
+        help='Provisioning priority group.',
+    )
     keyname: str = field(
         default=DEFAULT_KEYNAME,
         option='--keyname',
         metavar='NAME',
-        help='SSH key name.')
+        help='SSH key name.',
+    )
     user_data: dict[str, str] = field(
         default_factory=dict,
         option='--user-data',
         metavar='KEY=VALUE',
         help='Optional data to attach to guest.',
         multiple=True,
-        normalize=_normalize_user_data)
+        normalize=normalize_string_dict,
+    )
     kickstart: dict[str, str] = field(
         default_factory=dict,
         option='--kickstart',
         metavar='KEY=VALUE',
         help='Optional Beaker kickstart to use when provisioning the guest.',
         multiple=True,
-        normalize=_normalize_user_data)
+        normalize=normalize_string_dict,
+    )
 
     log_type: list[str] = field(
         default_factory=list,
@@ -191,10 +179,14 @@ class ArtemisGuestData(tmt.steps.provision.GuestSshData):
         choices=SUPPORTED_LOG_TYPES,
         help='Log types the guest must support.',
         multiple=True,
-        normalize=_normalize_log_type)
+        normalize=_normalize_log_type,
+    )
 
     # Provided by Artemis response
-    guestname: Optional[str] = None
+    guestname: Optional[str] = field(
+        default=None,
+        internal=True,
+    )
 
     # Timeouts and deadlines
     provision_timeout: int = field(
@@ -205,7 +197,8 @@ class ArtemisGuestData(tmt.steps.provision.GuestSshData):
              How long to wait for provisioning to complete,
              {DEFAULT_PROVISION_TIMEOUT} seconds by default.
              """,
-        normalize=tmt.utils.normalize_int)
+        normalize=tmt.utils.normalize_int,
+    )
     provision_tick: int = field(
         default=DEFAULT_PROVISION_TICK,
         option='--provision-tick',
@@ -214,7 +207,8 @@ class ArtemisGuestData(tmt.steps.provision.GuestSshData):
              How often check Artemis API for provisioning status,
              {DEFAULT_PROVISION_TICK} seconds by default.
              """,
-        normalize=tmt.utils.normalize_int)
+        normalize=tmt.utils.normalize_int,
+    )
     api_timeout: int = field(
         default=DEFAULT_API_TIMEOUT,
         option='--api-timeout',
@@ -223,7 +217,8 @@ class ArtemisGuestData(tmt.steps.provision.GuestSshData):
              How long to wait for API operations to complete,
              {DEFAULT_API_TIMEOUT} seconds by default.
              """,
-        normalize=tmt.utils.normalize_int)
+        normalize=tmt.utils.normalize_int,
+    )
     api_retries: int = field(
         default=DEFAULT_API_RETRIES,
         option='--api-retries',
@@ -232,7 +227,8 @@ class ArtemisGuestData(tmt.steps.provision.GuestSshData):
              How many attempts to use when talking to API,
              {DEFAULT_API_RETRIES} by default.
              """,
-        normalize=tmt.utils.normalize_int)
+        normalize=tmt.utils.normalize_int,
+    )
     api_retry_backoff_factor: int = field(
         default=DEFAULT_RETRY_BACKOFF_FACTOR,
         option='--api-retry-backoff-factor',
@@ -241,7 +237,8 @@ class ArtemisGuestData(tmt.steps.provision.GuestSshData):
              A factor for exponential API retry backoff,
             {DEFAULT_RETRY_BACKOFF_FACTOR} by default.
             """,
-        normalize=tmt.utils.normalize_int)
+        normalize=tmt.utils.normalize_int,
+    )
     # Artemis core already contains default values
     watchdog_dispatch_delay: Optional[int] = field(
         default=cast(Optional[int], None),
@@ -251,27 +248,30 @@ class ArtemisGuestData(tmt.steps.provision.GuestSshData):
              How long (seconds) before the guest "is-alive" watchdog is dispatched. The dispatch
              timer starts once the guest is successfully provisioned.
              """,
-        normalize=tmt.utils.normalize_optional_int)
+        normalize=tmt.utils.normalize_optional_int,
+    )
     watchdog_period_delay: Optional[int] = field(
         default=cast(Optional[int], None),
         option='--watchdog-period-delay',
         metavar='SECONDS',
         help='How often (seconds) check that the guest "is-alive".',
-        normalize=tmt.utils.normalize_optional_int)
+        normalize=tmt.utils.normalize_optional_int,
+    )
     skip_prepare_verify_ssh: bool = field(
         default=False,
         option='--skip-prepare-verify-ssh',
         is_flag=True,
-        help='If set, skip verifiction of SSH connection in prepare state.'
-        )
+        help='If set, skip verifiction of SSH connection in prepare state.',
+    )
     post_install_script: Optional[str] = field(
         default=None,
         option='--post-install-script',
         metavar='SCRIPT|URL',
-        help='If set, the script provided or fetched will be executed.')
+        help='If set, the script provided or fetched will be executed.',
+    )
 
 
-@dataclasses.dataclass
+@container
 class ProvisionArtemisData(ArtemisGuestData, tmt.steps.provision.ProvisionStepData):
     pass
 
@@ -284,12 +284,13 @@ class ArtemisProvisionError(ProvisionError):
     """
 
     def __init__(
-            self,
-            message: str,
-            response: Optional[requests.Response] = None,
-            request_data: Optional[dict[str, Any]] = None,
-            *args: Any,
-            **kwargs: Any) -> None:
+        self,
+        message: str,
+        response: Optional[requests.Response] = None,
+        request_data: Optional[dict[str, Any]] = None,
+        *args: Any,
+        **kwargs: Any,
+    ) -> None:
         if response is not None:
             message_components = [
                 message,
@@ -297,21 +298,18 @@ class ArtemisProvisionError(ProvisionError):
                 'Artemis API responded with unexpected code:',
                 '',
                 f'Request: {response.request.url}',
-                ]
+            ]
 
             if request_data is not None:
-                message_components += [
-                    '',
-                    dict_to_yaml(request_data)
-                    ]
+                message_components += ['', dict_to_yaml(request_data)]
 
             message_components += [
                 f'Response: {response.status_code} {response.reason}',
                 '',
                 dict_to_yaml(dict(response.headers)),
                 '',
-                dict_to_yaml(response.json())
-                ]
+                dict_to_yaml(response.json()),
+            ]
 
             message = '\n'.join(message_components)
 
@@ -326,8 +324,8 @@ GUEST_STATE_COLORS = {
     'promised': 'blue',
     'preparing': 'cyan',
     'cancelled': 'red',
-    'error': 'red'
-    }
+    'error': 'red',
+}
 
 
 # Type annotation for Artemis API `GET /guests/$guestname` response.
@@ -350,17 +348,17 @@ class ArtemisAPI:
                 500,  # Internal Server Error
                 502,  # Bad Gateway
                 503,  # Service Unavailable
-                504   # Gateway Timeout
-                ),
-            timeout=guest.api_timeout
-            )
+                504,  # Gateway Timeout
+            ),
+            timeout=guest.api_timeout,
+        )
 
     def query(
-            self,
-            path: str,
-            method: str = 'get',
-            request_kwargs: Optional[dict[str, Any]] = None
-            ) -> requests.Response:
+        self,
+        path: str,
+        method: str = 'get',
+        request_kwargs: Optional[dict[str, Any]] = None,
+    ) -> requests.Response:
         """
         Base helper for Artemis API queries.
 
@@ -388,15 +386,14 @@ class ArtemisAPI:
         if method == 'put':
             return self.http_session.put(url, **request_kwargs)
 
-        raise tmt.utils.GeneralError(
-            f'Unsupported Artemis API method.\n{method}')
+        raise tmt.utils.GeneralError(f'Unsupported Artemis API method.\n{method}')
 
     def create(
-            self,
-            path: str,
-            data: dict[str, Any],
-            request_kwargs: Optional[dict[str, Any]] = None
-            ) -> requests.Response:
+        self,
+        path: str,
+        data: dict[str, Any],
+        request_kwargs: Optional[dict[str, Any]] = None,
+    ) -> requests.Response:
         """
         Create - or request creation of - a resource.
 
@@ -412,11 +409,11 @@ class ArtemisAPI:
         return self.query(path, method='post', request_kwargs=request_kwargs)
 
     def inspect(
-            self,
-            path: str,
-            params: Optional[dict[str, Any]] = None,
-            request_kwargs: Optional[dict[str, Any]] = None
-            ) -> requests.Response:
+        self,
+        path: str,
+        params: Optional[dict[str, Any]] = None,
+        request_kwargs: Optional[dict[str, Any]] = None,
+    ) -> requests.Response:
         """
         Inspect a resource.
 
@@ -434,10 +431,10 @@ class ArtemisAPI:
         return self.query(path, request_kwargs=request_kwargs)
 
     def delete(
-            self,
-            path: str,
-            request_kwargs: Optional[dict[str, Any]] = None
-            ) -> requests.Response:
+        self,
+        path: str,
+        request_kwargs: Optional[dict[str, Any]] = None,
+    ) -> requests.Response:
         """
         Delete - or request removal of - a resource.
 
@@ -486,13 +483,15 @@ class GuestArtemis(tmt.GuestSsh):
     watchdog_dispatch_delay: Optional[int]
     watchdog_period_delay: Optional[int]
 
-    @cached_property
+    @functools.cached_property
     def api(self) -> ArtemisAPI:
         return ArtemisAPI(self)
 
     @property
     def is_ready(self) -> bool:
-        """ Detect the guest is ready or not """
+        """
+        Detect the guest is ready or not
+        """
 
         # FIXME: A more robust solution should be provided. Currently just
         #        return True if self.guest is not None
@@ -500,27 +499,24 @@ class GuestArtemis(tmt.GuestSsh):
 
     def _create(self) -> None:
         environment: dict[str, Any] = {
-            'hw': {
-                'arch': self.arch
-                },
-            'os': {
-                'compose': self.image
-                }
-            }
+            'hw': {'arch': self.arch},
+            'os': {'compose': self.image},
+        }
 
         if self.api_version >= "0.0.53":
             environment['kickstart'] = self.kickstart
 
         elif self.kickstart:
             raise ArtemisProvisionError(
-                f"API version '{self.api_version}' does not support kickstart.")
+                f"API version '{self.api_version}' does not support kickstart."
+            )
 
         data: dict[str, Any] = {
             'environment': environment,
             'keyname': self.keyname,
             'priority_group': self.priority_group,
-            'user_data': self.user_data
-            }
+            'user_data': self.user_data,
+        }
 
         if self.pool:
             environment['pool'] = self.pool
@@ -534,7 +530,8 @@ class GuestArtemis(tmt.GuestSsh):
 
         elif self.skip_prepare_verify_ssh:
             raise ArtemisProvisionError(
-                f"API version '{self.api_version}' does not support skip_prepare_verify_ssh.")
+                f"API version '{self.api_version}' does not support skip_prepare_verify_ssh."
+            )
 
         if self.api_version >= "0.0.56":
             if self.watchdog_dispatch_delay:
@@ -544,7 +541,8 @@ class GuestArtemis(tmt.GuestSsh):
 
         elif any([self.watchdog_dispatch_delay, self.watchdog_period_delay]):
             raise ArtemisProvisionError(
-                f"API version '{self.api_version}' does not support watchdog specification.")
+                f"API version '{self.api_version}' does not support watchdog specification."
+            )
 
         # TODO: snapshots
         # TODO: spot instance
@@ -556,12 +554,12 @@ class GuestArtemis(tmt.GuestSsh):
                         response = session.get(self.post_install_script)
 
                     if not response.ok:
-                        raise ArtemisProvisionError(
-                            "Failed to download the post-install script")
+                        raise ArtemisProvisionError("Failed to download the post-install script")
 
                 except requests.RequestException as exc:
                     raise ArtemisProvisionError(
-                        "Failed to download the post-install script") from exc
+                        "Failed to download the post-install script"
+                    ) from exc
 
                 post_install_script = response.text
             else:
@@ -574,15 +572,17 @@ class GuestArtemis(tmt.GuestSsh):
         if self.log_type:
             data['log_types'] = list({tuple(log.split('/', 1)) for log in self.log_type})
 
-        response = self.api.create('/guests/', data)
+        with tmt.utils.signals.PreventSignals(self._logger):
+            response = self.api.create('/guests/', data)
+
+            if response.status_code != 201:
+                raise ArtemisProvisionError(
+                    'Failed to create', response=response, request_data=data
+                )
 
-        if response.status_code == 201:
             self.info('guest', 'has been requested', 'green')
+            self.guestname = response.json()['guestname']
 
-        else:
-            raise ArtemisProvisionError('Failed to create', response=response, request_data=data)
-
-        self.guestname = response.json()['guestname']
         self.info('guestname', self.guestname, 'green')
 
         with UpdatableMessage('state', indent_level=self._level()) as progress_message:
@@ -595,8 +595,7 @@ class GuestArtemis(tmt.GuestSsh):
 
                 current = cast(GuestInspectType, response.json())
                 state = current['state']
-                state_color = GUEST_STATE_COLORS.get(
-                    state, GUEST_STATE_COLOR_DEFAULT)
+                state_color = GUEST_STATE_COLORS.get(state, GUEST_STATE_COLOR_DEFAULT)
 
                 progress_message.update(state, color=state_color)
 
@@ -610,8 +609,11 @@ class GuestArtemis(tmt.GuestSsh):
 
             try:
                 guest_info = tmt.utils.wait(
-                    self, get_new_state, datetime.timedelta(
-                        seconds=self.provision_timeout), tick=self.provision_tick)
+                    self,
+                    get_new_state,
+                    datetime.timedelta(seconds=self.provision_timeout),
+                    tick=self.provision_tick,
+                )
 
             except tmt.utils.WaitingTimedOutError:
                 # The provisioning chain has been already started, make sure we
@@ -620,7 +622,8 @@ class GuestArtemis(tmt.GuestSsh):
 
                 raise ArtemisProvisionError(
                     f'Failed to provision in the given amount '
-                    f'of time (--provision-timeout={self.provision_timeout}).')
+                    f'of time (--provision-timeout={self.provision_timeout}).'
+                )
 
         self.primary_address = self.topology_address = guest_info['address']
 
@@ -640,7 +643,9 @@ class GuestArtemis(tmt.GuestSsh):
         self.verbose('topology address', self.topology_address, 'green')
 
     def remove(self) -> None:
-        """ Remove the guest """
+        """
+        Remove the guest
+        """
 
         if self.guestname is None:
             return
@@ -658,9 +663,63 @@ class GuestArtemis(tmt.GuestSsh):
 
         else:
             self.info(
-                'guest',
-                f"Failed to remove, "
-                f"unhandled API response '{response.status_code}'.")
+                'guest', f"Failed to remove, unhandled API response '{response.status_code}'."
+            )
+
+    def reboot(
+        self,
+        hard: bool = False,
+        command: Optional[Union[Command, ShellScript]] = None,
+        timeout: Optional[int] = None,
+        tick: float = tmt.utils.DEFAULT_WAIT_TICK,
+        tick_increase: float = tmt.utils.DEFAULT_WAIT_TICK_INCREASE,
+    ) -> bool:
+        """
+        Reboot the guest, and wait for the guest to recover.
+
+        :param hard: if set, force the reboot. This may result in a loss of
+            data. The default of ``False`` will attempt a graceful reboot.
+        :param command: a command to run on the guest to trigger the reboot.
+        :param timeout: amount of time in which the guest must become available
+            again.
+        :param tick: how many seconds to wait between two consecutive attempts
+            of contacting the guest.
+        :param tick_increase: a multiplier applied to ``tick`` after every
+            attempt.
+        :returns: ``True`` if the reboot succeeded, ``False`` otherwise.
+        """
+
+        if hard:
+            if self.guestname is None:
+                raise ArtemisProvisionError("Cannot reboot - guest does not exist")
+
+            self.debug("Hard reboot using the Artemis API.")
+
+            def trigger_reboot() -> None:
+                response = self.api.query(f'/guests/{self.guestname}/reboot', method='post')
+                if response.status_code == 202:
+                    self.info('guest', 'reboot requested', 'green')
+                else:
+                    raise ArtemisProvisionError('Failed to reboot guest', response=response)
+
+            return self.perform_reboot(
+                trigger_reboot,
+                timeout=timeout,
+                tick=tick,
+                tick_increase=tick_increase,
+                fetch_boot_time=False,
+            )
+
+        actual_command = command or tmt.steps.DEFAULT_REBOOT_COMMAND
+        self.debug(f"Reboot using the command '{actual_command}'.")
+
+        return super().reboot(
+            hard=False,
+            command=actual_command,
+            timeout=timeout,
+            tick=tick,
+            tick_increase=tick_increase,
+        )
 
 
 @tmt.steps.provides_method('artemis')
@@ -668,6 +727,21 @@ class ProvisionArtemis(tmt.steps.provision.ProvisionPlugin[ProvisionArtemisData]
     """
     Provision guest using Artemis backend.
 
+    Reserve a machine using the Artemis service.
+    Users can specify many requirements, mostly regarding the
+    desired OS, RAM, disk size and more. Most of the HW specifications
+    defined in the :ref:`/spec/hardware` are supported. Including the
+    :ref:`/spec/plans/provision/kickstart`.
+
+    Artemis takes machines from AWS, OpenStack, Beaker or Azure.
+    By default, Artemis handles the selection of a cloud provider
+    to its best abilities and the required specification. However, it
+    is possible to specify the keyword ``pool`` and select the
+    desired cloud provider.
+
+    Artemis project:
+    https://gitlab.com/testing-farm/artemis
+
     Minimal configuration could look like this:
 
     .. code-block:: yaml
@@ -677,6 +751,12 @@ class ProvisionArtemis(tmt.steps.provision.ProvisionPlugin[ProvisionArtemisData]
             image: Fedora
             api-url: https://your-artemis.com/
 
+    .. note::
+
+        When used together with the Testing Farm infrastructure
+        some of the options from the first example below
+        will be filled for you by the Testing Farm service.
+
     .. note::
 
         The actual value of ``image`` depends on what images - or "composes" as
@@ -729,27 +809,17 @@ class ProvisionArtemis(tmt.steps.provision.ProvisionPlugin[ProvisionArtemisData]
     # Guest instance
     _guest = None
 
-    def go(self) -> None:
-        """ Provision the guest """
-        super().go()
+    def go(self, *, logger: Optional[tmt.log.Logger] = None) -> None:
+        """
+        Provision the guest
+        """
+
+        super().go(logger=logger)
 
         if self.data.api_version not in SUPPORTED_API_VERSIONS:
             raise ArtemisProvisionError(f"API version '{self.data.api_version}' not supported.")
 
-        try:
-            user_data = {
-                key.strip(): value.strip()
-                for key, value in (
-                    pair.split('=', 1)
-                    for pair in self.data.user_data
-                    )
-                }
-
-        except ValueError:
-            raise ArtemisProvisionError('Cannot parse user-data.')
-
         data = ArtemisGuestData.from_plugin(self)
-        data.user_data = user_data
 
         data.show(verbose=self.verbosity_level, logger=self._logger)
 
@@ -757,10 +827,7 @@ class ProvisionArtemis(tmt.steps.provision.ProvisionPlugin[ProvisionArtemisData]
             logger=self._logger,
             data=data,
             name=self.name,
-            parent=self.step)
+            parent=self.step,
+        )
         self._guest.start()
         self._guest.setup()
-
-    def guest(self) -> Optional[GuestArtemis]:
-        """ Return the provisioned guest """
-        return self._guest
